from __future__ import annotations

import ssl
from functools import cached_property
from typing import Any

import httpx

from infrahub import config
from infrahub.exceptions import HTTPServerError, HTTPServerSSLError, HTTPServerTimeoutError
from infrahub.log import get_logger
from infrahub.services.adapters.http import InfrahubHTTP

log = get_logger()


class HttpxAdapter(InfrahubHTTP):
    """The HttpxAdapter is a generic interface for InfrahubHTTP

    The class provides a way to send HTTP requests from Infrahub for example
    when sending webhooks, telemetry data or when communicating with SSO
    providers. The main purpose is to have a single location to manage
    configuration and error handling with regards to HTTP traffic and
    allow users to define configurations such as timeout, TLS options
    and eventually proxy settings in one location."""

    _settings: config.HTTPSettings | None = None

    @property
    def settings(self) -> config.HTTPSettings:
        if self._settings:
            return self._settings

        self._settings = config.SETTINGS.http
        return self._settings

    @cached_property
    def tls_context(self) -> ssl.SSLContext:
        return self.settings.get_tls_context()

    def verify_tls(self, verify: bool | None = None) -> bool | ssl.SSLContext:
        if verify is False:
            return False

        return self.tls_context

    async def _request(
        self,
        method: str,
        url: str,
        data: Any | None = None,
        json: Any | None = None,
        headers: dict[str, Any] | None = None,
        verify: bool | None = None,
    ) -> httpx.Response:
        """Returns an httpx.Response object or raises HTTPServerError or child classes."""
        params: dict[str, Any] = {}
        if data:
            params["data"] = data
        if json:
            params["json"] = json
        async with httpx.AsyncClient(verify=self.verify_tls(verify=verify)) as client:
            try:
                response = await client.request(
                    method=method,
                    url=url,
                    headers=headers,
                    timeout=self.settings.timeout,
                    **params,
                )
            except ssl.SSLCertVerificationError as exc:
                log.info(f"TLS verification failed for connection to {url}")
                raise HTTPServerSSLError(message=f"Unable to validate TLS certificate for connection to {url}") from exc
            except httpx.ReadTimeout as exc:
                log.info(f"Connection timed out when trying to reach {url}")
                raise HTTPServerTimeoutError(
                    message=f"Connection to {url} timed out after {self.settings.timeout}"
                ) from exc
            except httpx.RequestError as exc:
                # Catch all error from httpx
                log.warning(f"Unhandled HTTP error for {url} ({exc})")
                raise HTTPServerError(message=f"Unknown http error when connecting to {url}") from exc

        return response

    async def get(
        self,
        url: str,
        headers: dict[str, Any] | None = None,
    ) -> httpx.Response:
        return await self._request(
            method="get",
            url=url,
            headers=headers,
        )

    async def post(
        self,
        url: str,
        data: Any | None = None,
        json: Any | None = None,
        headers: dict[str, Any] | None = None,
        verify: bool | None = None,
    ) -> httpx.Response:
        return await self._request(method="post", url=url, data=data, json=json, headers=headers, verify=verify)



from infrahub.log import get_logger
from infrahub.message_bus import InfrahubMessage
from infrahub.message_bus.types import KVTTL
from infrahub.workers.dependencies import get_cache

log = get_logger()


async def set_check_status(message: InfrahubMessage, conclusion: str) -> None:
    if message.meta.validator_execution_id and message.meta.check_execution_id:
        key = f"validator_execution_id:{message.meta.validator_execution_id}:check_execution_id:{message.meta.check_execution_id}"
        cache = await get_cache()
        await cache.set(key=key, value=conclusion, expires=KVTTL.TWO_HOURS)
        log.debug("setting check status in cache", key=key, conclusion=conclusion)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generator

from infrahub import config
from infrahub.core.constants import GLOBAL_BRANCH_NAME, BranchSupportType, DiffAction, RelationshipStatus
from infrahub.core.query import Query, QueryType
from infrahub.core.timestamp import Timestamp

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.diff.model.field_specifiers_map import NodeFieldSpecifierMap
    from infrahub.database import InfrahubDatabase


class DiffQuery(Query):
    branch_names: list[str]
    diff_from: Timestamp
    diff_to: Timestamp
    type: QueryType = QueryType.READ

    def __init__(
        self,
        branch: Branch,
        diff_from: Timestamp | str = None,
        diff_to: Timestamp | str = None,
        **kwargs,
    ):
        """A diff is always in the context of a branch"""

        if not diff_from and branch.is_default:
            raise ValueError("diff_from is mandatory when the diff is on the main branch.")

        # If diff from hasn't been provided, we'll use the creation of the branch as the starting point
        if diff_from:
            self.diff_from = Timestamp(diff_from)
        else:
            self.diff_from = Timestamp(branch.created_at)

        # If Diff_to is not defined it will automatically select the current time.
        self.diff_to = Timestamp(diff_to)

        if self.diff_to < self.diff_from:
            raise ValueError("diff_to must be later than diff_from")

        self.branch_names = branch.get_branches_in_scope()

        super().__init__(branch, **kwargs)


class DiffCountChanges(Query):
    name = "diff_count_changes"
    type = QueryType.READ

    def __init__(
        self,
        branch_names: list[str],
        diff_from: Timestamp,
        diff_to: Timestamp,
        **kwargs,
    ):
        self.branch_names = branch_names
        self.diff_from = diff_from
        self.diff_to = diff_to
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        self.params = {
            "from_time": self.diff_from.to_string(),
            "to_time": self.diff_to.to_string(),
            "branch_names": self.branch_names,
        }
        query = """
        MATCH (p)-[diff_rel]-(q)
        WHERE any(l in labels(p) WHERE l in ["Node", "Attribute", "Relationship"])
        AND diff_rel.branch in $branch_names
        AND (
            (diff_rel.from >= $from_time AND diff_rel.from < $to_time)
            OR (diff_rel.to >= $to_time AND diff_rel.to < $to_time)
        )
        AND (p.branch_support = "aware" OR q.branch_support = "aware")
        WITH diff_rel.branch AS branch_name, count(*) AS num_changes
        """
        self.add_to_query(query=query)
        self.return_labels = ["branch_name", "num_changes"]

    def get_num_changes_by_branch(self) -> dict[str, int]:
        branch_count_map = {}
        for result in self.get_results():
            branch_name = str(result.get("branch_name"))
            try:
                count = int(result.get("num_changes"))
            except (TypeError, ValueError):
                count = 0
            branch_count_map[branch_name] = count
        for branch_name in self.branch_names:
            if branch_name not in branch_count_map:
                branch_count_map[branch_name] = 0
        return branch_count_map


async def get_num_changes_in_time_range_by_branch(
    branch_names: list[str],
    from_time: Timestamp,
    to_time: Timestamp,
    db: InfrahubDatabase,
) -> dict[str, int]:
    query = await DiffCountChanges.init(db=db, branch_names=branch_names, diff_from=from_time, diff_to=to_time)
    await query.execute(db=db)
    return query.get_num_changes_by_branch()


class DiffCalculationQuery(DiffQuery):
    type = QueryType.READ
    insert_limit = False

    def __init__(
        self,
        base_branch: Branch,
        diff_branch_from_time: Timestamp,
        current_node_field_specifiers: NodeFieldSpecifierMap | None = None,
        new_node_field_specifiers: NodeFieldSpecifierMap | None = None,
        **kwargs: Any,
    ):
        self.base_branch = base_branch
        self.diff_branch_from_time = diff_branch_from_time
        self.current_node_field_specifiers = current_node_field_specifiers
        self.new_node_field_specifiers = new_node_field_specifiers

        super().__init__(**kwargs)

    previous_base_path_query = """
WITH DISTINCT diff_path AS diff_path, has_more_data
CALL (diff_path) {
    WITH nodes(diff_path) AS d_nodes, relationships(diff_path) AS d_rels
    WITH d_rels[0] AS r_root, d_nodes[1] AS n, d_rels[1] AS r_node, d_nodes[2] AS attr_rel, d_rels[2] AS r_prop
    // -------------------------------------
    // add base branch paths before branched_from, if they exist
    // -------------------------------------
    WITH n, attr_rel, r_node, r_prop
    // 'base_n' instead of 'n' here to get previous value for node with a migrated kind/inheritance
    OPTIONAL MATCH latest_base_path = (:Root)<-[base_r_root:IS_PART_OF {branch: $base_branch_name}]
        -(base_n {uuid: n.uuid})-[base_r_node {branch: $base_branch_name}]
        -(attr_rel)-[base_r_prop {branch: $base_branch_name}]->(base_prop)
    WHERE type(base_r_node) = type(r_node)
    AND type(base_r_prop) = type(r_prop)
    AND [%(id_func)s(base_n), type(base_r_node)] <> [%(id_func)s(base_prop), type(base_r_prop)]
    AND all(
        r in relationships(latest_base_path)
        WHERE r.from < $branch_from_time
    )
    // ------------------------
    // special handling for nodes that had their kind updated,
    // the migration leaves two nodes with the same UUID linked to the same Relationship
    // ------------------------
    AND (
        base_n.uuid IS NULL OR base_prop.uuid IS NULL OR base_n.uuid <> base_prop.uuid
        OR type(base_r_node) <> "IS_RELATED" OR type(base_r_prop) <> "IS_RELATED"
    )
    WITH latest_base_path, base_r_root, base_r_node, base_r_prop
    // status="active" ordering is for tie-breaking edges added and deleted at the same time, we want the active one
    ORDER BY
        base_r_prop.from DESC,
        base_r_prop.status = "active" DESC,
        base_r_node.from DESC,
        base_r_node.status = "active" DESC,
        base_r_root.from DESC,
        base_r_root.status = "active" DESC
    LIMIT 1
    RETURN latest_base_path
}
    """
    relationship_peer_side_query = """
WITH diff_path, latest_base_path, has_more_data
UNWIND [diff_path, latest_base_path] AS penultimate_path
WITH DISTINCT penultimate_path, has_more_data
CALL (penultimate_path) {
    WITH nodes(penultimate_path) AS d_nodes, relationships(penultimate_path) AS d_rels
    WITH d_rels[0] AS r_root, d_nodes[1] AS n, d_rels[1] AS r_node, d_nodes[2] AS attr_rel, d_rels[2] AS r_prop
    // -------------------------------------
    // Add peer-side of any relationships to get the peer's ID
    // -------------------------------------
    WITH r_root, n, r_node, attr_rel, r_prop
    OPTIONAL MATCH peer_path = (
        (:Root)<-[peer_r_root:IS_PART_OF]-(n)-[peer_r_node:IS_RELATED]-(attr_rel:Relationship)-[r_peer:IS_RELATED]-(peer:Node)
    )
    WHERE type(r_prop) <> "IS_RELATED"
    AND %(id_func)s(peer_r_root) = %(id_func)s(r_root)
    AND %(id_func)s(peer_r_node) = %(id_func)s(r_node)
    AND [%(id_func)s(n), type(peer_r_node)] <> [%(id_func)s(peer), type(r_peer)]
    AND r_peer.from < $to_time
    // filter out paths where a base branch edge follows a branch edge
    AND (peer_r_node.branch = $base_branch_name OR r_peer.branch = $branch_name)
    // filter out paths where an active edge follows a deleted edge
    AND (peer_r_node.status = "active" OR r_peer.status = "deleted")
    // require adjacent edge pairs to have overlapping times, but only if on the same branch
    AND (
        peer_r_node.branch <> r_peer.branch
        OR peer_r_node.to IS NULL
        OR peer_r_node.to >= r_peer.from
    )
    // ------------------------
    // special handling for nodes that had their kind updated,
    // the migration leaves two nodes with the same UUID linked to the same Relationship
    // ------------------------
    AND (n.uuid IS NULL OR peer.uuid IS NULL OR n.uuid <> peer.uuid)
    WITH peer_path, r_peer, r_prop
    ORDER BY r_peer.branch = r_prop.branch DESC, r_peer.status = r_prop.status DESC, r_peer.from DESC, r_peer.status ASC
    LIMIT 1
    RETURN peer_path
}
WITH penultimate_path, peer_path, has_more_data
WITH reduce(
    diff_rel_paths = [], item IN [penultimate_path, peer_path] |
    CASE WHEN item IS NULL THEN diff_rel_paths ELSE diff_rel_paths + [item] END
) AS diff_rel_paths, has_more_data
// ------------------------
// make sure we still include has_more_data if diff_rel_paths is empty
// ------------------------
WITH CASE
    WHEN diff_rel_paths = [] THEN [NULL]
    ELSE diff_rel_paths
END AS diff_rel_paths, has_more_data
    """

    def get_previous_base_path_query(self, db: InfrahubDatabase) -> str:
        return self.previous_base_path_query % {"id_func": db.get_id_function_name()}

    def get_relationship_peer_side_query(self, db: InfrahubDatabase) -> str:
        return self.relationship_peer_side_query % {"id_func": db.get_id_function_name()}

    def get_params(self) -> dict[str, Any]:
        from_str = self.diff_from.to_string()
        return {
            "base_branch_name": self.base_branch.name,
            "branch_name": self.branch.name,
            "global_branch_name": GLOBAL_BRANCH_NAME,
            "branch_from_time": self.diff_branch_from_time.to_string(),
            "from_time": from_str,
            "to_time": self.diff_to.to_string(),
            "branch_local": BranchSupportType.LOCAL.value,
            "branch_aware": BranchSupportType.AWARE.value,
            "branch_agnostic": BranchSupportType.AGNOSTIC.value,
            "limit": self.limit or config.SETTINGS.database.query_size_limit,
            "offset": self.offset or 0,
        }


class DiffNodePathsQuery(DiffCalculationQuery):
    name = "diff_node_paths"

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        params_dict = self.get_params()
        self.params.update(params_dict)
        self.params.update(
            {
                "new_node_ids_list": self.new_node_field_specifiers.get_uuids_list()
                if self.new_node_field_specifiers
                else None,
                "current_node_ids_list": self.current_node_field_specifiers.get_uuids_list()
                if self.current_node_field_specifiers
                else None,
            }
        )
        nodes_path_query = """
// -------------------------------------
// Identify nodes added/removed on branch
// -------------------------------------
MATCH (q:Root)<-[diff_rel:IS_PART_OF {branch: $branch_name}]-(p:Node)
WHERE (
    ($new_node_ids_list IS NOT NULL AND p.uuid IN $new_node_ids_list)
    OR ($current_node_ids_list IS NOT NULL AND p.uuid IN $current_node_ids_list)
    OR ($new_node_ids_list IS NULL AND $current_node_ids_list IS NULL)
)
AND p.branch_support = $branch_aware
AND (
    (
        ($new_node_ids_list IS NOT NULL AND p.uuid IN $new_node_ids_list)
        AND (
            ($branch_from_time <= diff_rel.from < $to_time AND (diff_rel.to IS NULL OR diff_rel.to > $to_time))
            OR ($branch_from_time <= diff_rel.to < $to_time)
        )
    )
    OR (
        (
            ($current_node_ids_list IS NOT NULL AND p.uuid IN $current_node_ids_list)
            OR ($current_node_ids_list IS NULL AND $new_node_ids_list IS NULL)
        )
        AND (
            ($from_time <= diff_rel.from < $to_time AND (diff_rel.to IS NULL OR diff_rel.to > $to_time))
            OR ($from_time <= diff_rel.to < $to_time)
        )
    )
)
// -------------------------------------
// Limit the number of nodes
// -------------------------------------
WITH p, q, diff_rel, CASE
    WHEN $new_node_ids_list IS NOT NULL AND p.uuid IN $new_node_ids_list THEN $branch_from_time
    ELSE $from_time
END AS row_from_time
ORDER BY %(id_func)s(p) DESC
SKIP $offset
LIMIT $limit
// -------------------------------------
// Add flag to indicate if there is more data after this
// -------------------------------------
WITH collect([p, q, diff_rel, row_from_time]) AS limited_results
WITH limited_results, size(limited_results) = $limit AS has_more_data
UNWIND limited_results AS one_result
WITH one_result[0] AS p, one_result[1] AS q, one_result[2] AS diff_rel, one_result[3] AS row_from_time, has_more_data
// -------------------------------------
// Exclude nodes added then removed on branch within timeframe
// -------------------------------------
CALL (p, q, row_from_time) {
    OPTIONAL MATCH (q)<-[is_part_of:IS_PART_OF {branch: $branch_name}]-(p)
    WHERE row_from_time <= is_part_of.from < $to_time
    WITH DISTINCT is_part_of.status AS rel_status
    WITH collect(rel_status) AS rel_statuses
    RETURN ("active" IN rel_statuses AND "deleted" IN rel_statuses) AS intra_branch_update
}
WITH p, q, diff_rel, row_from_time, has_more_data, intra_branch_update
WHERE intra_branch_update = FALSE
// -------------------------------------
// Get every path on this branch under each node
// -------------------------------------
CALL (p, q, diff_rel, row_from_time) {
    OPTIONAL MATCH path = (
        (q)<-[top_diff_rel:IS_PART_OF]-(p)-[r_node]-(node)-[r_prop]-(prop)
    )
    WHERE %(id_func)s(diff_rel) = %(id_func)s(top_diff_rel)
    AND type(r_node) IN ["HAS_ATTRIBUTE", "IS_RELATED"]
    AND any(l in labels(node) WHERE l in ["Attribute", "Relationship"])
    AND node.branch_support IN [$branch_aware, $branch_agnostic]
    AND type(r_prop) IN ["IS_VISIBLE", "IS_PROTECTED", "HAS_SOURCE", "HAS_OWNER", "HAS_VALUE", "IS_RELATED"]
    AND any(l in labels(prop) WHERE l in ["Boolean", "Node", "AttributeValue"])
    AND (top_diff_rel.to IS NULL OR top_diff_rel.to >= r_node.from)
    AND (r_node.to IS NULL OR r_node.to >= r_prop.from)
    AND [%(id_func)s(p), type(r_node)] <> [%(id_func)s(prop), type(r_prop)]
    AND r_node.from < $to_time
    AND r_node.branch = top_diff_rel.branch
    AND r_node.status = top_diff_rel.status
    AND r_prop.from < $to_time
    AND r_prop.branch = top_diff_rel.branch
    AND r_prop.status = top_diff_rel.status
    // ------------------------
    // special handling for nodes that had their kind updated,
    // the migration leaves two nodes with the same UUID linked to the same Relationship
    // ------------------------
    AND (
        p.uuid IS NULL OR prop.uuid IS NULL OR p.uuid <> prop.uuid
        OR type(r_node) <> "IS_RELATED" OR type(r_prop) <> "IS_RELATED"
    )
    WITH path, node, prop, r_prop, r_node, type(r_node) AS rel_type, row_from_time
    // -------------------------------------
    // Exclude attributes/relationships added then removed on branch within timeframe
    // -------------------------------------
    CALL (p, rel_type, node, row_from_time) {
        OPTIONAL MATCH (p)-[rel_to_check {branch: $branch_name}]-(node)
        WHERE row_from_time <= rel_to_check.from < $to_time
        AND type(rel_to_check) = rel_type
        WITH DISTINCT rel_to_check.status AS rel_status
        WITH collect(rel_status) AS rel_statuses
        RETURN ("active" IN rel_statuses AND "deleted" IN rel_statuses) AS intra_branch_update
    }
    WITH path, node, prop, r_prop, r_node, intra_branch_update
    WHERE intra_branch_update = FALSE
    WITH path, node, prop, r_prop, r_node
    ORDER BY
        %(id_func)s(node),
        %(id_func)s(prop),
        r_prop.from DESC,
        r_node.from DESC
    WITH node, prop, type(r_prop) AS r_prop_type, type(r_node) AS r_node_type, head(collect(path)) AS diff_path
    RETURN diff_path
}
""" % {"id_func": db.get_id_function_name()}
        self.add_to_query(nodes_path_query)
        self.add_to_query(self.get_previous_base_path_query(db=db))
        self.add_to_query(self.get_relationship_peer_side_query(db=db))
        self.add_to_query("UNWIND diff_rel_paths AS diff_path")
        self.return_labels = ["DISTINCT diff_path AS diff_path", "has_more_data"]


class DiffFieldPathsQuery(DiffCalculationQuery):
    name = "diff_field_paths"

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        params_dict = self.get_params()
        self.params.update(params_dict)

        self.params.update(
            {
                "current_node_ids_list": self.current_node_field_specifiers.get_uuids_list()
                if self.current_node_field_specifiers
                else None,
                "new_node_ids_list": self.new_node_field_specifiers.get_uuids_list()
                if self.new_node_field_specifiers
                else None,
                "current_node_field_specifiers_map": self.current_node_field_specifiers.get_uuid_field_names_map()
                if self.current_node_field_specifiers is not None
                else None,
                "new_node_field_specifiers_map": self.new_node_field_specifiers.get_uuid_field_names_map()
                if self.new_node_field_specifiers is not None
                else None,
            }
        )
        fields_path_query = """
// -------------------------------------
// Identify attributes/relationships added/removed on branch
// -------------------------------------
MATCH (root:Root)<-[r_root:IS_PART_OF]-(p:Node)-[diff_rel {branch: $branch_name}]-(q)
// simple filters to start
WHERE type(diff_rel) IN ["HAS_ATTRIBUTE", "IS_RELATED"]
AND ("Attribute" IN labels(q) OR "Relationship" IN labels(q))
AND r_root.branch IN [$branch_name, $base_branch_name, $global_branch_name]
AND q.branch_support = $branch_aware
AND r_root.status = "active"
AND r_root.from <= diff_rel.from
AND (r_root.to IS NULL OR diff_rel.branch <> r_root.branch OR r_root.to >= diff_rel.from)
// node ID and field name filtering first pass
AND (
    (
        $current_node_ids_list IS NOT NULL
        AND p.uuid IN $current_node_ids_list
        AND q.name IN $current_node_field_specifiers_map[p.uuid]
    ) OR (
        $new_node_ids_list IS NOT NULL
        AND p.uuid IN $new_node_ids_list
        AND q.name IN $new_node_field_specifiers_map[p.uuid]
    ) OR (
        $new_node_ids_list IS NULL
        AND $current_node_ids_list IS NULL
    )
)
// node ID and field name filtering second pass
AND (
    // time-based filters for nodes already included in the diff or fresh changes
    (
        (
            (
                $current_node_ids_list IS NOT NULL
                AND p.uuid IN $current_node_ids_list
                AND q.name IN $current_node_field_specifiers_map[p.uuid]
            )
            OR ($current_node_ids_list IS NULL AND $new_node_ids_list IS NULL)
        )
        AND (r_root.from < $from_time OR p.branch_support = $branch_agnostic)
        AND (
            ($from_time <= diff_rel.from < $to_time AND (diff_rel.to IS NULL OR diff_rel.to > $to_time))
            OR ($from_time <= diff_rel.to < $to_time)
        )
    )
    // time-based filters for new nodes
    OR (
        (
            $new_node_ids_list IS NOT NULL
            AND p.uuid IN $new_node_ids_list
            AND q.name IN $new_node_field_specifiers_map[p.uuid]
        )
        AND (r_root.from < $branch_from_time OR p.branch_support = $branch_agnostic)
        AND (
            ($branch_from_time <= diff_rel.from < $to_time AND (diff_rel.to IS NULL OR diff_rel.to > $to_time))
            OR ($branch_from_time <= diff_rel.to < $to_time)
        )
    )
)
// -------------------------------------
// Limit the number of paths
// -------------------------------------
WITH root, r_root, p, diff_rel, q
ORDER BY r_root.from, p.uuid, q.uuid, diff_rel.branch, diff_rel.from
SKIP $offset
LIMIT $limit
// -------------------------------------
// Add flag to indicate if there is more data after this
// -------------------------------------
WITH collect([root, r_root, p, diff_rel, q]) AS limited_results
WITH limited_results, size(limited_results) = $limit AS has_more_data
UNWIND limited_results AS one_result
WITH one_result[0] AS root, one_result[1] AS r_root, one_result[2] AS p, one_result[3] AS diff_rel, one_result[4] AS q, has_more_data
// -------------------------------------
// Add correct from_time for row
// -------------------------------------
WITH root, r_root, p, diff_rel, q, has_more_data, CASE
    WHEN
        $new_node_ids_list IS NOT NULL
        AND p.uuid IN $new_node_ids_list
        AND q.name IN $new_node_field_specifiers_map[p.uuid]
    THEN $branch_from_time
    ELSE $from_time
END AS row_from_time
// -------------------------------------
// Exclude attributes/relationship under nodes deleted on this branch in the timeframe
// because those were all handled above at the node level
// -------------------------------------
CALL (root, p, row_from_time) {
    OPTIONAL MATCH (root)<-[r_root_deleted:IS_PART_OF {branch: $branch_name}]-(p)
    WHERE row_from_time <= r_root_deleted.from < $to_time
    WITH r_root_deleted
    ORDER BY r_root_deleted.status DESC
    LIMIT 1
    RETURN COALESCE(r_root_deleted.status = "deleted", FALSE) AS node_deleted
}
WITH root, r_root, p, diff_rel, q, has_more_data, row_from_time, node_deleted
WHERE node_deleted = FALSE
// -------------------------------------
// Exclude relationships added and deleted within the timeframe
// -------------------------------------
WITH root, r_root, p, diff_rel, q, has_more_data, row_from_time, type(diff_rel) AS rel_type
CALL (p, rel_type, q, row_from_time) {
    OPTIONAL MATCH (p)-[rel_to_check {branch: $branch_name}]-(q)
    WHERE row_from_time <= rel_to_check.from < $to_time
    AND type(rel_to_check) = rel_type
    WITH DISTINCT rel_to_check.status AS rel_status
    WITH collect(rel_status) AS rel_statuses
    RETURN ("active" IN rel_statuses AND "deleted" IN rel_statuses) AS intra_branch_update
}
WITH root, r_root, p, diff_rel, q, has_more_data, row_from_time, intra_branch_update
WHERE intra_branch_update = FALSE
// -------------------------------------
// Get every path on this branch under each attribute/relationship
// -------------------------------------
CALL (root, r_root, p, diff_rel, q) {
    OPTIONAL MATCH path = (
        (root:Root)<-[mid_r_root:IS_PART_OF]-(p)-[mid_diff_rel]-(q)-[r_prop]-(prop)
    )
    WHERE %(id_func)s(mid_r_root) =  %(id_func)s(r_root)
    AND %(id_func)s(mid_diff_rel) =  %(id_func)s(diff_rel)
    AND type(r_prop) IN ["IS_VISIBLE", "IS_PROTECTED", "HAS_SOURCE", "HAS_OWNER", "HAS_VALUE", "IS_RELATED"]
    AND any(l in labels(prop) WHERE l in ["Boolean", "Node", "AttributeValue"])
    AND r_prop.from < $to_time AND r_prop.branch = mid_diff_rel.branch
    AND (mid_diff_rel.to IS NULL OR mid_diff_rel.to >= r_prop.from)
    AND [%(id_func)s(p), type(mid_diff_rel)] <> [%(id_func)s(prop), type(r_prop)]
    // exclude paths where an active edge is below a deleted edge
    AND (mid_diff_rel.status = "active" OR r_prop.status = "deleted")
    // ------------------------
    // special handling for nodes that had their kind updated,
    // the migration leaves two nodes with the same UUID linked to the same Relationship
    // ------------------------
    AND (
        p.uuid IS NULL OR prop.uuid IS NULL OR p.uuid <> prop.uuid
        OR type(mid_diff_rel) <> "IS_RELATED" OR type(r_prop) <> "IS_RELATED"
    )
    WITH path, prop, r_prop, mid_r_root
    ORDER BY
        type(r_prop),
        mid_r_root.branch = mid_diff_rel.branch DESC,
        (mid_diff_rel.status = r_prop.status AND mid_diff_rel.branch = r_prop.branch) DESC,
        r_prop.from DESC,
        mid_r_root.from DESC
    WITH prop, type(r_prop) AS type_r_prop, head(collect(path)) AS latest_prop_path
    RETURN latest_prop_path
}
// -------------------------------------
// Exclude properties added and deleted within the timeframe
// -------------------------------------
WITH q, nodes(latest_prop_path)[3] AS prop, type(relationships(latest_prop_path)[2]) AS rel_type, latest_prop_path, has_more_data, row_from_time
CALL (q, rel_type, prop, row_from_time) {
    OPTIONAL MATCH (q)-[rel_to_check {branch: $branch_name}]-(prop)
    WHERE row_from_time <= rel_to_check.from < $to_time
    AND type(rel_to_check) = rel_type
    WITH DISTINCT rel_to_check.status AS rel_status
    WITH collect(rel_status) AS rel_statuses
    RETURN ("active" IN rel_statuses AND "deleted" IN rel_statuses) AS intra_branch_update
}
WITH latest_prop_path AS diff_path, has_more_data, intra_branch_update
WHERE intra_branch_update = FALSE
        """ % {"id_func": db.get_id_function_name()}
        self.add_to_query(fields_path_query)
        self.add_to_query(self.get_previous_base_path_query(db=db))
        self.add_to_query(self.get_relationship_peer_side_query(db=db))
        self.add_to_query("UNWIND diff_rel_paths AS diff_path")
        self.return_labels = ["DISTINCT diff_path AS diff_path", "has_more_data"]


class DiffPropertyPathsQuery(DiffCalculationQuery):
    name = "diff_property_paths"

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        params_dict = self.get_params()
        self.params.update(params_dict)

        self.params.update(
            {
                "current_node_ids_list": self.current_node_field_specifiers.get_uuids_list()
                if self.current_node_field_specifiers
                else None,
                "new_node_ids_list": self.new_node_field_specifiers.get_uuids_list()
                if self.new_node_field_specifiers
                else None,
                "current_node_field_specifiers_map": self.current_node_field_specifiers.get_uuid_field_names_map()
                if self.current_node_field_specifiers is not None
                else None,
                "new_node_field_specifiers_map": self.new_node_field_specifiers.get_uuid_field_names_map()
                if self.new_node_field_specifiers is not None
                else None,
            }
        )
        properties_path_query = """
// -------------------------------------
// Identify properties added/removed on branch
// -------------------------------------
MATCH diff_rel_path = (root:Root)<-[r_root:IS_PART_OF]-(n:Node)-[r_node]-(p)-[diff_rel {branch: $branch_name}]->(q)
WHERE p.branch_support = $branch_aware
AND any(l in labels(p) WHERE l in ["Attribute", "Relationship"])
AND type(diff_rel) IN ["IS_VISIBLE", "IS_PROTECTED", "HAS_SOURCE", "HAS_OWNER", "HAS_VALUE"]
AND any(l in labels(q) WHERE l in ["Boolean", "Node", "AttributeValue"])
AND type(r_node) IN ["HAS_ATTRIBUTE", "IS_RELATED"]
// node ID and field name filtering first pass
AND (
    (
        $current_node_ids_list IS NOT NULL
        AND n.uuid IN $current_node_ids_list
        AND p.name IN $current_node_field_specifiers_map[n.uuid]
    ) OR (
        $new_node_ids_list IS NOT NULL
        AND n.uuid IN $new_node_ids_list
        AND p.name IN $new_node_field_specifiers_map[n.uuid]
    ) OR (
        $new_node_ids_list IS NULL
        AND $current_node_ids_list IS NULL
    )
)
// node ID and field name filtering second pass
AND (
    // time-based filters for nodes already included in the diff or fresh changes
    (
        (
            (
                $current_node_ids_list IS NOT NULL
                AND n.uuid IN $current_node_ids_list
                AND p.name IN $current_node_field_specifiers_map[n.uuid]
            )
            OR ($current_node_ids_list IS NULL AND $new_node_ids_list IS NULL)
        )
        AND (
            ($from_time <= diff_rel.from < $to_time AND (diff_rel.to IS NULL OR diff_rel.to > $to_time))
            OR ($from_time <= diff_rel.to < $to_time)
        )
        // skip paths where nodes/attrs/rels are updated after $from_time, those are handled in other queries
        AND (
            r_root.from <= $from_time AND (r_root.to IS NULL OR r_root.branch <> diff_rel.branch OR r_root.to <= $from_time)
            AND r_node.from <= $from_time AND (r_node.to IS NULL OR r_node.branch <> diff_rel.branch OR r_node.to <= $from_time)
        )
    )
    // time-based filters for new nodes
    OR (
        (
            $new_node_ids_list IS NOT NULL
            AND n.uuid IN $new_node_ids_list
            AND p.name IN $new_node_field_specifiers_map[n.uuid]
        )
        AND (
            ($branch_from_time <= diff_rel.from < $to_time AND (diff_rel.to IS NULL OR diff_rel.to > $to_time))
            OR ($branch_from_time <= diff_rel.to < $to_time)
        )
        // skip paths where nodes/attrs/rels are updated after $branch_from_time, those are handled in other queries
        AND (
            r_root.from <= $branch_from_time AND (r_root.to IS NULL OR r_root.branch <> diff_rel.branch OR r_root.to <= $branch_from_time)
            AND r_node.from <= $branch_from_time AND (r_node.to IS NULL OR r_node.branch <> diff_rel.branch OR r_node.to <= $branch_from_time)
        )
    )
)
// ------------------------
// special handling for nodes that had their kind updated,
// the migration leaves two nodes with the same UUID linked to the same Relationship
// ------------------------
AND (
    n.uuid IS NULL OR q.uuid IS NULL OR n.uuid <> q.uuid
    OR type(r_node) <> "IS_RELATED" OR type(diff_rel) <> "IS_RELATED"
)
AND ALL(
    r_pair IN [[r_root, r_node], [r_node, diff_rel]]
    // filter out paths where a base branch edge follows a branch edge
    WHERE ((r_pair[0]).branch = $base_branch_name OR (r_pair[1]).branch = $branch_name)
    // filter out paths where an active edge follows a deleted edge
    AND ((r_pair[0]).status = "active" OR (r_pair[1]).status = "deleted")
    // filter out paths where an earlier from time follows a later from time
    AND (r_pair[0]).from <= (r_pair[1]).from
    // if both are deleted, then the deeper edge must have been deleted first
    AND ((r_pair[0]).status = "active" OR (r_pair[1]).status = "active" OR (r_pair[0]).from >= (r_pair[1].from))
    AND (
        (r_pair[0]).status = (r_pair[1]).status
        OR (
            (r_pair[0]).from <= (r_pair[1]).from
            AND ((r_pair[0]).to IS NULL OR (r_pair[0]).to >= (r_pair[1]).from)
        )
    )
    // require adjacent edge pairs to have overlapping times, but only if on the same branch
    AND (
        (r_pair[0]).branch <> (r_pair[1]).branch
        OR (r_pair[0]).to IS NULL
        OR (r_pair[0]).to >= (r_pair[1]).from
    )
)
AND [%(id_func)s(n), type(r_node)] <> [%(id_func)s(q), type(diff_rel)]
// -------------------------------------
// Limit the number of paths
// -------------------------------------
WITH diff_rel_path, r_root, n, r_node, p, diff_rel
ORDER BY r_root.from, n.uuid, p.uuid, type(diff_rel), diff_rel.branch, diff_rel.from
SKIP $offset
LIMIT $limit
// -------------------------------------
// Add flag to indicate if there is more data after this
// -------------------------------------
WITH collect([diff_rel_path, r_root, n, r_node, p, diff_rel]) AS limited_results
WITH limited_results, size(limited_results) = $limit AS has_more_data
UNWIND limited_results AS one_result
WITH one_result[0] AS diff_rel_path, one_result[1] AS r_root, one_result[2] AS n,
    one_result[3] AS r_node, one_result[4] AS p, one_result[5] AS diff_rel, has_more_data
// -------------------------------------
// Add correct from_time for row
// -------------------------------------
WITH diff_rel_path, r_root, n, r_node, p, diff_rel, has_more_data, CASE
    WHEN
        $new_node_ids_list IS NOT NULL
        AND n.uuid IN $new_node_ids_list
        AND p.name IN $new_node_field_specifiers_map[n.uuid]
    THEN $branch_from_time
    ELSE $from_time
END AS row_from_time
WITH diff_rel_path, r_root, n, r_node, p, diff_rel, has_more_data, row_from_time
ORDER BY
    %(id_func)s(n) DESC,
    %(id_func)s(p) DESC,
    type(diff_rel),
    r_node.branch = diff_rel.branch DESC,
    r_root.branch = diff_rel.branch DESC,
    diff_rel.from DESC,
    r_node.from DESC,
    r_root.from DESC
WITH n, p, row_from_time, diff_rel, diff_rel_path, has_more_data
CALL (n, p, row_from_time){
    // -------------------------------------
    // Exclude properties under nodes and attributes/relationships deleted
    // on this branch in the timeframe because those were all handled above
    // -------------------------------------
    CALL (n, row_from_time) {
        OPTIONAL MATCH (root:Root)<-[r_root_deleted:IS_PART_OF {branch: $branch_name}]-(n)
        WHERE r_root_deleted.from < $to_time
        WITH r_root_deleted
        ORDER BY r_root_deleted.status DESC
        LIMIT 1
        RETURN COALESCE(r_root_deleted.status = "deleted", FALSE) AS node_deleted
    }
    WITH node_deleted
    CALL (n, p, row_from_time) {
        OPTIONAL MATCH (n)-[r_node_deleted {branch: $branch_name}]-(p)
        WHERE row_from_time <= r_node_deleted.from < $to_time
        AND type(r_node_deleted) IN ["HAS_ATTRIBUTE", "IS_RELATED"]
        WITH r_node_deleted
        ORDER BY r_node_deleted.status DESC
        LIMIT 1
        RETURN COALESCE(r_node_deleted.status = "deleted", FALSE) AS field_deleted
    }
    RETURN node_deleted OR field_deleted AS node_or_field_deleted
}
WITH n, p, diff_rel, diff_rel_path, has_more_data, node_or_field_deleted
WHERE node_or_field_deleted = FALSE
WITH n, p, type(diff_rel) AS drt, head(collect(diff_rel_path)) AS diff_path, has_more_data
        """ % {"id_func": db.get_id_function_name()}
        self.add_to_query(properties_path_query)
        self.add_to_query(self.get_previous_base_path_query(db=db))
        self.add_to_query(self.get_relationship_peer_side_query(db=db))
        self.add_to_query("UNWIND diff_rel_paths AS diff_path")
        self.return_labels = ["DISTINCT diff_path AS diff_path", "has_more_data"]


@dataclass
class MigratedKindNode:
    uuid: str
    kind: str
    db_id: str
    from_time: Timestamp
    action: DiffAction
    has_more_data: bool


class DiffMigratedKindNodesQuery(DiffCalculationQuery):
    name = "diff_migrated_kind_nodes_query"

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        params_dict = self.get_params()
        self.params.update(params_dict)
        migrated_kind_nodes_query = """
// -------------------------------------
// Identify nodes added/removed on branch in the time frame
// -------------------------------------
MATCH (:Root)<-[diff_rel:IS_PART_OF {branch: $branch_name}]-(n:Node)
WHERE (
    ($from_time <= diff_rel.from < $to_time AND (diff_rel.to IS NULL OR diff_rel.to > $to_time))
    OR ($from_time <= diff_rel.to < $to_time)
)
AND n.branch_support = $branch_aware
WITH DISTINCT n.uuid AS node_uuid, %(id_func)s(n) AS db_id
WITH node_uuid, count(*) AS num_nodes_with_uuid
WHERE num_nodes_with_uuid > 1
// -------------------------------------
// Limit the number of nodes
// -------------------------------------
WITH node_uuid
ORDER BY node_uuid
SKIP $offset
LIMIT $limit
WITH collect(node_uuid) AS node_uuids
WITH node_uuids, size(node_uuids) = $limit AS has_more_data
MATCH (:Root)<-[diff_rel:IS_PART_OF {branch: $branch_name}]-(n:Node)
WHERE n.uuid IN node_uuids
AND (
    ($from_time <= diff_rel.from < $to_time AND (diff_rel.to IS NULL OR diff_rel.to > $to_time))
    OR ($from_time <= diff_rel.to < $to_time)
)
// -------------------------------------
// Ignore node created and deleted on this branch
// -------------------------------------
CALL (n) {
    OPTIONAL MATCH (:Root)<-[diff_rel:IS_PART_OF {branch: $branch_name}]-(n)
    WITH diff_rel
    ORDER BY diff_rel.from ASC
    WITH collect(diff_rel.status) AS statuses
    RETURN statuses = ["active", "deleted"] AS intra_branch_update
}
WITH n.uuid AS uuid, n.kind AS kind, %(id_func)s(n) AS db_id, diff_rel.from_time AS from_time, diff_rel.status AS status, has_more_data
WHERE intra_branch_update = FALSE
        """ % {"id_func": db.get_id_function_name()}
        self.add_to_query(query=migrated_kind_nodes_query)
        self.return_labels = [
            "uuid",
            "kind",
            "db_id",
            "from_time",
            "status",
            "has_more_data",
        ]

    def get_migrated_kind_nodes(self) -> Generator[MigratedKindNode, None, None]:
        for result in self.get_results():
            yield MigratedKindNode(
                uuid=result.get_as_type("uuid", return_type=str),
                kind=result.get_as_type("kind", return_type=str),
                db_id=result.get_as_type("db_id", return_type=str),
                from_time=result.get_as_type("from_time", return_type=Timestamp),
                action=DiffAction.REMOVED
                if result.get_as_type("status", return_type=str).lower() == RelationshipStatus.DELETED.value
                else DiffAction.ADDED,
                has_more_data=result.get_as_type("has_more_data", bool),
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

from infrahub.core import registry
from infrahub.core.account import GlobalPermission
from infrahub.core.constants import GlobalPermissions, PermissionDecision
from infrahub.exceptions import PermissionDeniedError
from infrahub.permissions.constants import GLOBAL_PERMISSION_DENIAL_MESSAGE, PermissionDecisionFlag

if TYPE_CHECKING:
    from infrahub.auth import AccountSession
    from infrahub.core.account import ObjectPermission
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase
    from infrahub.permissions.types import AssignedPermissions

__all__ = ["PermissionManager"]


class PermissionManager:
    wildcard_values = ["*"]
    wildcard_actions = ["any"]

    def __init__(self, account_session: AccountSession) -> None:
        self.account_session = account_session
        self.permissions: AssignedPermissions = {"global_permissions": [], "object_permissions": []}

    async def load_permissions(self, db: InfrahubDatabase, branch: Branch) -> None:
        """Load permissions from the configured backends into memory."""
        for permission_backend in registry.permission_backends:
            backend_permissions = await permission_backend.load_permissions(
                db=db, branch=branch, account_session=self.account_session
            )
            self.permissions["global_permissions"].extend(backend_permissions["global_permissions"])
            self.permissions["object_permissions"].extend(backend_permissions["object_permissions"])

    def _compute_specificity(self, permission: ObjectPermission) -> int:
        """Return how specific a permission is."""
        specificity = 0
        if permission.namespace not in self.wildcard_values:
            specificity += 1
        if permission.name not in self.wildcard_values:
            specificity += 1
        if permission.action not in self.wildcard_actions:
            specificity += 1
        if not permission.decision & PermissionDecisionFlag.ALLOW_ALL:
            specificity += 1
        return specificity

    def is_super_admin(self) -> bool:
        return self.resolve_global_permission(
            permission_to_check=GlobalPermission(
                action=GlobalPermissions.SUPER_ADMIN, decision=PermissionDecision.ALLOW_ALL
            ),
        )

    def report_object_permission(self, namespace: str, name: str, action: str) -> PermissionDecisionFlag:
        """Given a set of permissions, return the permission decision for a given kind and action."""
        highest_specificity: int = -1
        combined_decision = PermissionDecisionFlag.DENY

        for permission in self.permissions["object_permissions"]:
            if (
                permission.namespace in [namespace, *self.wildcard_values]
                and permission.name in [name, *self.wildcard_values]
                and permission.action in [action, *self.wildcard_actions]
            ):
                permission_decision = PermissionDecisionFlag(value=permission.decision)
                # Compute the specifity of a permission to keep the decision of the most specific if two or more permissions overlap
                specificity = self._compute_specificity(permission=permission)
                if specificity > highest_specificity:
                    combined_decision = permission_decision
                    highest_specificity = specificity
                elif specificity == highest_specificity and permission_decision != PermissionDecisionFlag.DENY:
                    combined_decision |= permission_decision

        return combined_decision

    def resolve_object_permission(self, permission_to_check: ObjectPermission) -> bool:
        """Compute the permissions and check if the one provided is granted."""
        required_decision = PermissionDecisionFlag(value=permission_to_check.decision)
        combined_decision = self.report_object_permission(
            namespace=permission_to_check.namespace, name=permission_to_check.name, action=permission_to_check.action
        )

        return combined_decision & required_decision == required_decision

    def resolve_global_permission(self, permission_to_check: GlobalPermission) -> bool:
        """Tell if a global permission is granted."""
        grant_permission = False

        for permission in self.permissions["global_permissions"]:
            if permission.action == permission_to_check.action:
                # Early exit on deny as deny preempt allow
                if permission.decision == PermissionDecisionFlag.DENY:
                    return False
                grant_permission = True

        return grant_permission

    def has_permission(self, permission: GlobalPermission | ObjectPermission) -> bool:
        """Tell if a permission is granted given the permissions loaded in memory."""
        is_super_admin = self.is_super_admin()

        if isinstance(permission, GlobalPermission):
            return self.resolve_global_permission(permission_to_check=permission) or is_super_admin

        return self.resolve_object_permission(permission_to_check=permission) or is_super_admin

    def has_permissions(self, permissions: Sequence[GlobalPermission | ObjectPermission]) -> bool:
        """Same as `has_permission` but for multiple permissions, return `True` only if all permissions are granted."""
        return all(self.has_permission(permission=permission) for permission in permissions)

    def raise_for_permission(self, permission: GlobalPermission | ObjectPermission, message: str = "") -> None:
        """Same as `has_permission` but raise a `PermissionDeniedError` if the permission is not granted."""
        if self.has_permission(permission=permission):
            return

        if not message:
            if isinstance(permission, GlobalPermission) and permission.action in GLOBAL_PERMISSION_DENIAL_MESSAGE:
                message = GLOBAL_PERMISSION_DENIAL_MESSAGE[permission.action]
            else:
                message = f"You do not have the following permission: {permission!s}"

        raise PermissionDeniedError(message=message)

    def raise_for_permissions(
        self, permissions: Sequence[GlobalPermission | ObjectPermission], message: str = ""
    ) -> None:
        """Same as `has_permissions` but raise a `PermissionDeniedError` if any of the permissions is not granted."""
        if self.has_permissions(permissions=permissions):
            return

        if not message:
            message = f"You do not have one of the following permissions: {' | '.join([str(p) for p in permissions])}"

        raise PermissionDeniedError(message=message)



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub_sdk.batch import InfrahubBatch
from prefect import flow, task
from prefect.cache_policies import NONE
from prefect.logging import get_run_logger

from infrahub.core.branch import Branch  # noqa: TC001
from infrahub.core.path import SchemaPath  # noqa: TC001
from infrahub.core.schema import GenericSchema, NodeSchema
from infrahub.core.validators.aggregated_checker import AggregatedConstraintChecker
from infrahub.core.validators.model import SchemaConstraintValidatorRequest, SchemaViolation
from infrahub.dependencies.registry import get_component_registry
from infrahub.workers.dependencies import get_database
from infrahub.workflows.utils import add_tags

from .models.validate_migration import SchemaValidateMigrationData, SchemaValidatorPathResponseData

if TYPE_CHECKING:
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.database import InfrahubDatabase


@flow(name="schema_validate_migrations", flow_run_name="Validate schema migrations", persist_result=True)
async def schema_validate_migrations(message: SchemaValidateMigrationData) -> list[SchemaValidatorPathResponseData]:
    batch = InfrahubBatch(return_exceptions=True)
    log = get_run_logger()
    await add_tags(branches=[message.branch.name])

    if not message.constraints:
        log.info("No constaint to validate")
        return []

    log.info(f"{len(message.constraints)} constraint(s) to validate")
    # NOTE this task is a good candidate to add a progress bar
    for constraint in message.constraints:
        schema = message.schema_branch.get(name=constraint.path.schema_kind)
        if not isinstance(schema, GenericSchema | NodeSchema):
            continue
        batch.add(
            task=schema_path_validate,
            branch=message.branch,
            constraint_name=constraint.constraint_name,
            node_schema=schema,
            schema_path=constraint.path,
            schema_branch=message.schema_branch,
            database=await get_database(),
        )

    results = [result async for _, result in batch.execute()]
    return results


@task(  # type: ignore[arg-type]
    name="schema-path-validate",
    task_run_name="Validate schema path {constraint_name} in {branch.name}",
    description="Validate if a given migration is compatible with the existing data",
    retries=3,
    cache_policy=NONE,
)
async def schema_path_validate(
    branch: Branch,
    constraint_name: str,
    node_schema: NodeSchema | GenericSchema,
    schema_path: SchemaPath,
    schema_branch: SchemaBranch,
    database: InfrahubDatabase,
) -> SchemaValidatorPathResponseData:
    async with database.start_session(read_only=True) as db:
        constraint_request = SchemaConstraintValidatorRequest(
            branch=branch,
            constraint_name=constraint_name,
            node_schema=node_schema,
            schema_path=schema_path,
            schema_branch=schema_branch,
        )

        component_registry = get_component_registry()
        aggregated_constraint_checker = await component_registry.get_component(
            AggregatedConstraintChecker, db=db, branch=branch
        )
        try:
            violations = await aggregated_constraint_checker.run_constraints(constraint_request)
        except Exception as exc:
            violation = SchemaViolation(
                node_id="unknown",
                node_kind=node_schema.kind,
                display_label=f"Error validating {constraint_name} on {node_schema.kind}",
                full_display_label=f"Error validating {constraint_name} on {node_schema.kind}",
                message=str(exc),
            )
            violations = [violation]

        return SchemaValidatorPathResponseData(
            violations=violations, constraint_name=constraint_name, schema_path=schema_path
        )



from typing import Any

from infrahub.core.constants import InfrahubKind

from .constants import BranchScope, MemberAction, MemberUpdate, RelationshipMatch, ValueMatch
from .models import (
    CoreAction,
    CoreGeneratorAction,
    CoreGroupAction,
    CoreGroupTriggerRule,
    CoreNodeTriggerAttributeMatch,
    CoreNodeTriggerMatch,
    CoreNodeTriggerRelationshipMatch,
    CoreNodeTriggerRule,
    CoreTriggerRule,
)


def parse_trigger_rule_response(data: dict[str, Any]) -> list[CoreTriggerRule]:
    rules: list[CoreTriggerRule] = []
    if kind := data.get(InfrahubKind.TRIGGERRULE):
        if edges := kind.get("edges"):
            for edge in edges:
                if rule := _parse_graphql_node(edge["node"]):
                    rules.append(rule)

    return rules


def _parse_graphql_node(data: dict[str, Any]) -> CoreTriggerRule | None:
    typename = data.get("__typename")
    name = data["name"]["value"]
    active = data["active"]["value"]
    branch_scope = BranchScope.from_value(value=data["branch_scope"]["value"])
    action = _parse_graphql_action_response(data=data["action"]["node"])
    match typename:
        case "CoreGroupTriggerRule":
            member_update = MemberUpdate.from_value(data["member_update"]["value"])
            group_id = data["group"]["node"]["id"]
            group_kind = data["group"]["node"]["__typename"]
            return CoreGroupTriggerRule(
                name=name,
                branch_scope=branch_scope,
                action=action,
                member_update=member_update,
                group_id=group_id,
                group_kind=group_kind,
                active=active,
            )
        case "CoreNodeTriggerRule":
            node_kind = data["node_kind"]["value"]
            mutation_action = data["mutation_action"]["value"]
            matches = _parse_node_trigger_matches(data=data["matches"]["edges"])
            return CoreNodeTriggerRule(
                name=name,
                branch_scope=branch_scope,
                action=action,
                node_kind=node_kind,
                mutation_action=mutation_action,
                matches=matches,
                active=active,
            )
    return None


def _parse_graphql_action_response(data: dict[str, Any]) -> CoreAction:
    typename = data["__typename"]
    match typename:
        case "CoreGeneratorAction":
            generator_id = data["generator"]["node"]["id"]
            return CoreGeneratorAction(generator_id=generator_id)
        case "CoreGroupAction":
            member_action = MemberAction.from_value(data["member_action"]["value"])
            group_id = data["group"]["node"]["id"]
            return CoreGroupAction(member_action=member_action, group_id=group_id)

    raise NotImplementedError(f"{typename} is not a valid CoreAction")


def _parse_node_trigger_matches(data: list[dict[str, Any]]) -> list[CoreNodeTriggerMatch]:
    matches: list[CoreNodeTriggerMatch] = []
    for entry in data:
        node = entry["node"]
        typename = node["__typename"]
        match typename:
            case "CoreNodeTriggerAttributeMatch":
                matches.append(
                    CoreNodeTriggerAttributeMatch(
                        attribute_name=node["attribute_name"]["value"],
                        value=node["value"]["value"],
                        value_previous=node["value_previous"]["value"],
                        value_match=ValueMatch.from_value(value=node["value_match"]["value"]),
                    )
                )
            case "CoreNodeTriggerRelationshipMatch":
                matches.append(
                    CoreNodeTriggerRelationshipMatch(
                        relationship_name=node["relationship_name"]["value"],
                        modification_type=RelationshipMatch.from_value(node["modification_type"]["value"]),
                        peer=node["peer"]["value"],
                    )
                )

    return matches



import logging
import os
import time
from contextlib import asynccontextmanager
from functools import partial
from pathlib import Path
from typing import AsyncGenerator, Awaitable, Callable

from asgi_correlation_id import CorrelationIdMiddleware
from asgi_correlation_id.context import correlation_id
from fastapi import FastAPI, Request, Response
from fastapi.logger import logger
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.responses import RedirectResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from infrahub_sdk.exceptions import TimestampFormatError
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
from opentelemetry.trace import Span
from starlette_exporter import PrometheusMiddleware, handle_metrics

from infrahub import __version__, config
from infrahub.api import router as api
from infrahub.api.exception_handlers import generic_api_exception_handler
from infrahub.components import ComponentType
from infrahub.constants.environment import INSTALLATION_TYPE
from infrahub.core.initialization import initialization
from infrahub.dependencies.registry import build_component_registry
from infrahub.exceptions import Error, ValidationError
from infrahub.graphql.api.endpoints import router as graphql_router
from infrahub.lock import initialize_lock
from infrahub.log import clear_log_context, get_logger, set_log_data
from infrahub.middleware import InfrahubCORSMiddleware
from infrahub.services import InfrahubServices
from infrahub.trace import add_span_exception, configure_trace, get_traceid
from infrahub.worker import WORKER_IDENTITY
from infrahub.workers.dependencies import (
    get_cache,
    get_component,
    get_database,
    get_installation_type,
    get_message_bus,
    get_workflow,
    set_component_type,
)

CURRENT_DIRECTORY = Path(__file__).parent.resolve()


async def app_initialization(application: FastAPI, enable_scheduler: bool = True) -> None:
    config.SETTINGS.initialize_and_exit()
    _validate_feature_selection(configuration=config.SETTINGS.active_settings)

    # Initialize trace
    if config.SETTINGS.trace.enable:
        configure_trace(
            service="infrahub-server",
            version=__version__,
            exporter_type=config.SETTINGS.trace.exporter_type,
            exporter_endpoint=config.SETTINGS.trace.exporter_endpoint,
            exporter_protocol=config.SETTINGS.trace.exporter_protocol,
        )

    component_type = ComponentType.API_SERVER
    set_component_type(component_type=component_type)

    # Initialize database Driver and load local registry
    database = application.state.db = await get_database()

    build_component_registry()

    workflow = get_workflow()
    message_bus = await get_message_bus()
    cache = await get_cache()
    component = await get_component()
    service = await InfrahubServices.new(
        cache=cache,
        database=database,
        message_bus=message_bus,
        workflow=workflow,
        component=component,
        component_type=component_type,
    )
    initialize_lock(service=service)
    # We must initialize DB after initialize lock and initialize lock depends on cache initialization
    async with application.state.db.start_session() as db:
        await initialization(db=db, add_database_indexes=True)

    application.state.service = service
    application.state.response_delay = config.SETTINGS.miscellaneous.response_delay
    if enable_scheduler:
        await service.scheduler.start_schedule()


async def shutdown(application: FastAPI) -> None:
    await application.state.service.shutdown()
    await application.state.db.close()


@asynccontextmanager
async def lifespan(application: FastAPI) -> AsyncGenerator:
    await app_initialization(application)
    yield
    await shutdown(application)


app = FastAPI(
    title="Infrahub",
    version=__version__,
    lifespan=lifespan,
    openapi_url="/api/openapi.json",
    docs_url=None,
    redoc_url=None,
)


def server_request_hook(span: Span, scope: dict) -> None:  # noqa: ARG001
    if span and span.is_recording():
        span.set_attribute("worker", WORKER_IDENTITY)


FastAPIInstrumentor().instrument_app(app, excluded_urls=".*/metrics", server_request_hook=server_request_hook)

FRONTEND_DIRECTORY = Path(os.environ.get("INFRAHUB_FRONTEND_DIRECTORY", "frontend/app")).resolve()
FRONTEND_ASSET_DIRECTORY = FRONTEND_DIRECTORY / "dist" / "assets"
FRONTEND_FAVICONS_DIRECTORY = FRONTEND_DIRECTORY / "dist" / "favicons"

DOCS_DIRECTORY = Path(os.environ.get("INFRAHUB_DOCS_DIRECTORY", Path("docs").resolve()))
DOCS_BUILD_DIRECTORY = DOCS_DIRECTORY / "build"

log = get_logger()
gunicorn_logger = logging.getLogger("gunicorn.error")
logger.handlers = gunicorn_logger.handlers

app.include_router(api)

templates = Jinja2Templates(directory=FRONTEND_DIRECTORY / "dist")


@app.middleware("http")
async def logging_middleware(request: Request, call_next: Callable[[Request], Awaitable[Response]]) -> Response:
    clear_log_context()
    request_id = correlation_id.get()

    set_log_data(key="request_id", value=request_id)
    set_log_data(key="app", value="infrahub.api")
    set_log_data(key="worker", value=WORKER_IDENTITY)

    trace_id = get_traceid()
    if trace_id:
        set_log_data(key="trace_id", value=trace_id)

    response = await call_next(request)
    return response


@app.middleware("http")
async def add_process_time_header(request: Request, call_next: Callable[[Request], Awaitable[Response]]) -> Response:
    start_time = time.time()
    response = await call_next(request)
    process_time = time.time() - start_time
    response.headers["X-Process-Time"] = str(process_time)
    return response


@app.middleware("http")
async def add_telemetry_span_exception(
    request: Request, call_next: Callable[[Request], Awaitable[Response]]
) -> Response:
    try:
        return await call_next(request)
    except Exception as exc:
        add_span_exception(exc)
        raise


app.add_middleware(CorrelationIdMiddleware)
app.add_middleware(
    PrometheusMiddleware,
    app_name="infrahub",
    group_paths=True,
    prefix="infrahub",
    buckets=[0.1, 0.25, 0.5],
    skip_paths=["/health"],
)
app.add_middleware(InfrahubCORSMiddleware)
app.add_middleware(GZipMiddleware, minimum_size=100_000)

app.add_exception_handler(Error, generic_api_exception_handler)
app.add_exception_handler(TimestampFormatError, partial(generic_api_exception_handler, http_code=400))
app.add_exception_handler(ValidationError, partial(generic_api_exception_handler, http_code=400))

app.add_route(path="/metrics", route=handle_metrics)
app.include_router(graphql_router)

app.mount("/api-static", StaticFiles(directory=CURRENT_DIRECTORY / "api" / "static"), name="static")

if FRONTEND_ASSET_DIRECTORY.exists() and FRONTEND_ASSET_DIRECTORY.is_dir():
    app.mount("/assets", StaticFiles(directory=FRONTEND_ASSET_DIRECTORY), "assets")


if FRONTEND_FAVICONS_DIRECTORY.exists() and FRONTEND_FAVICONS_DIRECTORY.is_dir():
    app.mount("/favicons", StaticFiles(directory=FRONTEND_FAVICONS_DIRECTORY), "favicons")


if DOCS_BUILD_DIRECTORY.exists() and DOCS_BUILD_DIRECTORY.is_dir():
    app.mount("/docs", StaticFiles(directory=DOCS_BUILD_DIRECTORY, html=True, check_dir=True), name="infrahub-docs")


@app.get("/docs", include_in_schema=False)
async def documentation() -> RedirectResponse:
    return RedirectResponse("/docs/")


@app.get("/{rest_of_path:path}", include_in_schema=False)
async def react_app(req: Request, rest_of_path: str) -> Response:  # noqa: ARG001
    return templates.TemplateResponse("index.html", {"request": req})


def _validate_feature_selection(configuration: config.Settings) -> None:
    if configuration.enterprise_features and not configuration.dev.allow_enterprise_configuration:
        installation_type = get_installation_type()
        if installation_type == INSTALLATION_TYPE:
            raise ValidationError(
                f"Enterprise features [{','.join(configuration.enterprise_features)}] are not supported when running Infrahub 'community'."
            )



from __future__ import annotations

from copy import deepcopy
from typing import Any

from pydantic import Field, field_validator, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Self

from .constants import InfrahubClientMode
from .playback import JSONPlayback
from .recorder import JSONRecorder, NoRecorder, Recorder, RecorderType
from .types import AsyncRequester, InfrahubLoggers, RequesterTransport, SyncRequester
from .utils import get_branch, is_valid_url


class ProxyMountsConfig(BaseSettings):
    model_config = SettingsConfigDict(populate_by_name=True)
    http: str | None = Field(
        default=None,
        description="Proxy for HTTP requests",
        alias="http://",
        validation_alias="INFRAHUB_PROXY_MOUNTS_HTTP",
    )
    https: str | None = Field(
        default=None,
        description="Proxy for HTTPS requests",
        alias="https://",
        validation_alias="INFRAHUB_PROXY_MOUNTS_HTTPS",
    )

    @property
    def is_set(self) -> bool:
        return self.http is not None or self.https is not None


class ConfigBase(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_", validate_assignment=True)
    address: str = Field(default="http://localhost:8000", description="The URL to use when connecting to Infrahub.")
    api_token: str | None = Field(default=None, description="API token for authentication against Infrahub.")
    echo_graphql_queries: bool = Field(
        default=False, description="If set the GraphQL query and variables will be echoed to the screen"
    )
    username: str | None = Field(default=None, description="Username for accessing Infrahub", min_length=1)
    password: str | None = Field(default=None, description="Password for accessing Infrahub", min_length=1)
    default_branch: str = Field(
        default="main", description="Default branch to target if not specified for each request."
    )
    default_branch_from_git: bool = Field(
        default=False,
        description="Indicates if the default Infrahub branch to target should come from the active branch in the local Git repository.",
    )
    identifier: str | None = Field(default=None, description="Tracker identifier")
    insert_tracker: bool = Field(default=False, description="Insert a tracker on queries to the server")
    max_concurrent_execution: int = Field(default=5, description="Max concurrent execution in batch mode")
    mode: InfrahubClientMode = Field(default=InfrahubClientMode.DEFAULT, description="Default mode for the client")
    pagination_size: int = Field(default=50, description="Page size for queries to the server")
    retry_delay: int = Field(default=5, description="Number of seconds to wait until attempting a retry.")
    retry_on_failure: bool = Field(default=False, description="Retry operation in case of failure")
    max_retry_duration: int = Field(
        default=300, description="Maximum duration until we stop attempting to retry if enabled."
    )
    schema_converge_timeout: int = Field(
        default=60, description="Number of seconds to wait for schema to have converged"
    )
    timeout: int = Field(default=60, description="Default connection timeout in seconds")
    transport: RequesterTransport = Field(
        default=RequesterTransport.HTTPX, description="Set an alternate transport using a predefined option"
    )
    proxy: str | None = Field(default=None, description="Proxy address")
    proxy_mounts: ProxyMountsConfig = Field(default=ProxyMountsConfig(), description="Proxy mounts configuration")
    update_group_context: bool = Field(default=False, description="Update GraphQL query groups")
    tls_insecure: bool = Field(
        default=False,
        description="""
    Indicates if TLS certificates are verified.
    Enabling this option will disable: CA verification, expiry date verification, hostname verification).
    Can be useful to test with self-signed certificates.""",
    )
    tls_ca_file: str | None = Field(default=None, description="File path to CA cert or bundle in PEM format")

    @model_validator(mode="before")
    @classmethod
    def validate_credentials_input(cls, values: dict[str, Any]) -> dict[str, Any]:
        has_username = "username" in values
        has_password = "password" in values
        if has_username != has_password:
            raise ValueError("Both 'username' and 'password' needs to be set")
        return values

    @model_validator(mode="before")
    @classmethod
    def set_transport(cls, values: dict[str, Any]) -> dict[str, Any]:
        if values.get("transport") == RequesterTransport.JSON:
            playback = JSONPlayback()
            if "requester" not in values:
                values["requester"] = playback.async_request
            if "sync_requester" not in values:
                values["sync_requester"] = playback.sync_request

        return values

    @model_validator(mode="before")
    @classmethod
    def validate_mix_authentication_schemes(cls, values: dict[str, Any]) -> dict[str, Any]:
        if values.get("password") and values.get("api_token"):
            raise ValueError("Unable to combine password with token based authentication")
        return values

    @field_validator("address")
    @classmethod
    def validate_address(cls, value: str) -> str:
        if is_valid_url(value):
            return value.rstrip("/")

        raise ValueError("The configured address is not a valid url")

    @model_validator(mode="after")
    def validate_proxy_config(self) -> Self:
        if self.proxy and self.proxy_mounts.is_set:
            raise ValueError("'proxy' and 'proxy_mounts' are mutually exclusive")
        return self

    @property
    def default_infrahub_branch(self) -> str:
        branch: str | None = None
        if not self.default_branch_from_git:
            branch = self.default_branch

        return get_branch(branch=branch)

    @property
    def password_authentication(self) -> bool:
        return bool(self.username)


class Config(ConfigBase):
    recorder: RecorderType = Field(default=RecorderType.NONE, description="Select builtin recorder for later replay.")
    custom_recorder: Recorder = Field(
        default_factory=NoRecorder.default, description="Provides a way to record responses from the Infrahub API"
    )
    requester: AsyncRequester | None = None
    sync_requester: SyncRequester | None = None
    log: Any | None = None

    @property
    def logger(self) -> InfrahubLoggers:
        # We expect the log to adhere to the definitions defined by the InfrahubLoggers object
        # When using structlog the logger doesn't expose the expected methods by looking at the
        # object to pydantic rejects them. This is a workaround to allow structlog to be used
        # as a logger
        return self.log  # type: ignore

    @model_validator(mode="before")
    @classmethod
    def set_custom_recorder(cls, values: dict[str, Any]) -> dict[str, Any]:
        if values.get("recorder") == RecorderType.NONE and "custom_recorder" not in values:
            values["custom_recorder"] = NoRecorder()
        elif values.get("recorder") == RecorderType.JSON and "custom_recorder" not in values:
            values["custom_recorder"] = JSONRecorder()
        return values

    def clone(self, branch: str | None = None) -> Config:
        config: dict[str, Any] = {
            "default_branch": branch or self.default_branch,
            "recorder": self.recorder,
            "custom_recorder": self.custom_recorder,
            "requester": self.requester,
            "sync_requester": self.sync_requester,
            "log": self.log,
        }
        covered_keys = list(config.keys())
        for field in Config.model_fields.keys():
            if field not in covered_keys:
                config[field] = deepcopy(getattr(self, field))

        return Config(**config)



import platform
from typing import Any

import psutil


def get_system_stats() -> dict[str, Any]:
    """
    Gather and return key system statistics about CPU, memory and platform
    Returns a dictionary with system information
    """
    # CPU information
    cpu_freq = psutil.cpu_freq()
    cpu_stats = {
        "cpu_count_physical": psutil.cpu_count(logical=False),
        "cpu_count_logical": psutil.cpu_count(logical=True),
        "cpu_freq_current": float(f"{cpu_freq.current:.2f}") if cpu_freq else None,  # MHz
        "cpu_freq_min": float(f"{cpu_freq.min:.2f}") if cpu_freq else None,  # MHz
        "cpu_freq_max": float(f"{cpu_freq.max:.2f}") if cpu_freq else None,  # MHz
        "cpu_percent": psutil.cpu_percent(interval=1, percpu=False),
    }

    # Memory information
    memory = psutil.virtual_memory()
    memory_stats = {
        "total_memory": float(f"{memory.total / (1024**3):.2f}"),  # GB
        "available_memory": float(f"{memory.available / (1024**3):.2f}"),  # GB
        "memory_percent_used": memory.percent,
        "memory_used": float(f"{memory.used / (1024**3):.2f}"),  # GB
    }

    # Platform information
    platform_stats = {
        "system": platform.system(),
        "platform_release": platform.release(),
        "platform_version": platform.version(),
        "architecture": platform.machine(),
        "processor": platform.processor(),
    }

    return {"cpu": cpu_stats, "memory": memory_stats, "platform": platform_stats}



from typing import ClassVar

from pydantic import Field

from .constants import EVENT_NAMESPACE
from .models import InfrahubEvent


class ValidatorEvent(InfrahubEvent):
    """Event generated when an validator within a pipeline has started."""

    node_id: str = Field(..., description="The ID of the validator")
    kind: str = Field(..., description="The kind of the validator")
    proposed_change_id: str = Field(..., description="The ID of the proposed change")

    def get_resource(self) -> dict[str, str]:
        return {
            "prefect.resource.id": self.node_id,
            "infrahub.node.kind": self.kind,
            "infrahub.node.id": self.node_id,
            "infrahub.branch.name": self.meta.context.branch.name,
        }

    def get_related(self) -> list[dict[str, str]]:
        related = super().get_related()
        related.append(
            {
                "prefect.resource.id": self.proposed_change_id,
                "prefect.resource.role": "infrahub.related.node",
                "infrahub.node.kind": "CoreProposedChange",
            }
        )

        return related


class ValidatorStartedEvent(ValidatorEvent):
    """Event generated when an validator within a pipeline has started."""

    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.validator.started"
    infrahub_node_kind_event: ClassVar[bool] = True


class ValidatorPassedEvent(ValidatorEvent):
    """Event generated when an validator within a pipeline has completed successfully."""

    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.validator.passed"
    infrahub_node_kind_event: ClassVar[bool] = True


class ValidatorFailedEvent(ValidatorEvent):
    """Event generated when an validator within a pipeline has completed successfully."""

    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.validator.failed"
    infrahub_node_kind_event: ClassVar[bool] = True



from typing import Any, Generator

from neo4j.graph import Node as Neo4jNode

from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase

from ..model.path import TrackingId


class EnrichedDiffAllConflictsQuery(Query):
    name = "enriched_diff_all_conflicts"
    type = QueryType.READ

    def __init__(
        self, diff_branch_name: str, tracking_id: TrackingId | None = None, diff_id: str | None = None, **kwargs: Any
    ) -> None:
        super().__init__(**kwargs)
        if (diff_id is None and tracking_id is None) or (diff_id and tracking_id):
            raise ValueError("EnrichedDiffAllConflictsQuery requires one and only one of `tracking_id` or `diff_id`")
        self.diff_branch_name = diff_branch_name
        self.tracking_id = tracking_id
        self.diff_id = diff_id
        if self.tracking_id is None and self.diff_id is None:
            raise RuntimeError("tracking_id or diff_id is required")

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {
            "diff_branch_name": self.diff_branch_name,
            "diff_id": self.diff_id,
            "tracking_id": self.tracking_id.serialize() if self.tracking_id else None,
        }
        query = """
MATCH (root:DiffRoot)
WHERE (root.is_merged IS NULL OR root.is_merged <> TRUE)
AND (
    ($diff_id IS NOT NULL AND root.uuid = $diff_id)
    OR ($tracking_id IS NOT NULL AND root.tracking_id = $tracking_id AND root.diff_branch = $diff_branch_name)
)
CALL (root) {
    MATCH (root)-[:DIFF_HAS_NODE]->(node:DiffNode)-[:DIFF_HAS_CONFLICT]->(node_conflict:DiffConflict)
    RETURN node.path_identifier AS path_identifier, node_conflict AS conflict
    UNION
    MATCH (root)-[:DIFF_HAS_NODE]->(node:DiffNode)-[:DIFF_HAS_ATTRIBUTE]->(:DiffAttribute)
        -[:DIFF_HAS_PROPERTY]->(property:DiffProperty)-[:DIFF_HAS_CONFLICT]->(attr_property_conflict:DiffConflict)
    RETURN property.path_identifier AS path_identifier, attr_property_conflict AS conflict
    UNION
    MATCH (root)-[:DIFF_HAS_NODE]->(node:DiffNode)-[:DIFF_HAS_RELATIONSHIP]->(:DiffRelationship)
        -[:DIFF_HAS_ELEMENT]->(element:DiffRelationshipElement)-[:DIFF_HAS_CONFLICT]->(rel_element_conflict:DiffConflict)
    RETURN element.path_identifier AS path_identifier, rel_element_conflict AS conflict
    UNION
    MATCH (root)-[:DIFF_HAS_NODE]->(node:DiffNode)-[:DIFF_HAS_RELATIONSHIP]->(:DiffRelationship)
        -[:DIFF_HAS_ELEMENT]->(:DiffRelationshipElement)-[:DIFF_HAS_PROPERTY]->(property:DiffProperty)
        -[:DIFF_HAS_CONFLICT]->(rel_property_conflict:DiffConflict)
    RETURN property.path_identifier AS path_identifier, rel_property_conflict AS conflict
}
"""
        self.return_labels = ["path_identifier", "conflict"]
        self.add_to_query(query=query)

    def get_conflict_paths_and_nodes(self) -> Generator[tuple[str, Neo4jNode], None, None]:
        for result in self.get_results():
            yield (result.get_as_type("path_identifier", str), result.get_node("conflict"))



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import GraphMigration, MigrationResult

from ...query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class CleanupOrphanedRelationshipsQuery(Query):
    name = "cleanup_orphaned_relationships"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
MATCH (rel:Relationship)-[:IS_RELATED]-(peer:Node)
WITH DISTINCT rel, peer.uuid AS p_uuid
WITH rel, count(*) AS num_peers
WHERE num_peers < 2
DETACH DELETE rel
        """
        self.add_to_query(query)


class Migration035(GraphMigration):
    """
    Remove Relationship vertices that only have a single peer
    """

    name: str = "035_clean_up_orphaned_relationships"
    minimum_version: int = 34
    queries: Sequence[type[Query]] = [CleanupOrphanedRelationshipsQuery]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, db: InfrahubDatabase) -> MigrationResult:
        # overrides parent class to skip transaction in case there are a lot of relationships to delete
        return await self.do_execute(db=db)



import os

from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
    OTLPSpanExporter as GRPCSpanExporter,
)
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter as HTTPSpanExporter,
)
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, ConsoleSpanExporter, SpanExporter
from opentelemetry.trace import StatusCode

from infrahub.worker import WORKER_IDENTITY


def get_current_span_with_context() -> trace.Span:
    return trace.get_current_span()


def get_traceid() -> str | None:
    current_span = get_current_span_with_context()
    trace_id = current_span.get_span_context().trace_id
    if trace_id == 0:
        return None
    return hex(trace_id)


def set_span_status(status_code: StatusCode) -> None:
    current_span = get_current_span_with_context()
    if current_span.is_recording():
        status = StatusCode(status_code)
        current_span.set_status(status)
        # current_span.set_attribute("status_code", status)


def set_span_data(key: str, value: str) -> None:
    current_span = get_current_span_with_context()
    if current_span.is_recording():
        current_span.set_attribute(key, value)


def add_span_event(event_name: str, event_attributes: dict) -> None:
    current_span = get_current_span_with_context()
    if current_span.is_recording():
        current_span.add_event(event_name, event_attributes)


def add_span_exception(exception: Exception) -> None:
    set_span_status(StatusCode.ERROR)
    current_span = get_current_span_with_context()
    if current_span.is_recording():
        current_span.record_exception(exception)


def create_tracer_provider(
    service: str,
    version: str,
    exporter_type: str,
    exporter_endpoint: str | None = None,
    exporter_protocol: str | None = None,
) -> TracerProvider:
    # Create a BatchSpanProcessor exporter based on the type
    if exporter_type == "console":
        exporter: SpanExporter = ConsoleSpanExporter()
    elif exporter_type == "otlp":
        if not exporter_endpoint:
            raise ValueError("Exporter type is set to otlp but endpoint is not set")
        if exporter_protocol == "http/protobuf":
            exporter = HTTPSpanExporter(endpoint=exporter_endpoint)
        elif exporter_protocol == "grpc":
            exporter = GRPCSpanExporter(endpoint=exporter_endpoint)
    else:
        raise ValueError("Exporter type unsupported by Infrahub")

    extra_attributes = {}
    if os.getenv("OTEL_RESOURCE_ATTRIBUTES"):
        extra_attributes = dict(attr.split("=") for attr in os.getenv("OTEL_RESOURCE_ATTRIBUTES", "").split(","))

    # Resource can be required for some backends, e.g. Jaeger
    resource = Resource(
        attributes={
            "service.name": service,
            "service.version": version,
            "worker.id": WORKER_IDENTITY,
            **extra_attributes,
        }
    )
    span_processor = BatchSpanProcessor(exporter)
    tracer_provider = TracerProvider(resource=resource)
    tracer_provider.add_span_processor(span_processor)

    return tracer_provider


def configure_trace(
    service: str,
    version: str,
    exporter_type: str,
    exporter_endpoint: str | None = None,
    exporter_protocol: str | None = None,
) -> None:
    # Create a trace provider with the exporter
    tracer_provider = create_tracer_provider(
        service=service,
        version=version,
        exporter_type=exporter_type,
        exporter_endpoint=exporter_endpoint,
        exporter_protocol=exporter_protocol,
    )

    # Register the trace provider
    trace.set_tracer_provider(tracer_provider)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Field, Int, List, NonNull, ObjectType, String

from infrahub.core.query.relationship import RelationshipGetByIdentifierQuery
from infrahub.graphql.field_extractor import extract_graphql_fields
from infrahub.graphql.types import RelationshipNode

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.graphql.initialization import GraphqlContext


class Relationships(ObjectType):
    edges = List(of_type=NonNull(RelationshipNode), required=True)
    count = Int(required=True)

    @staticmethod
    async def resolve(
        root: dict,  # noqa: ARG004
        info: GraphQLResolveInfo,
        ids: list[str],
        limit: int = 10,
        offset: int = 0,
        excluded_namespaces: list[str] | None = None,
    ) -> dict[str, Any]:
        graphql_context: GraphqlContext = info.context

        fields = extract_graphql_fields(info)
        excluded_namespaces = excluded_namespaces or []

        response: dict[str, Any] = {"edges": [], "count": None}

        async with graphql_context.db.start_session(read_only=True) as db:
            query = await RelationshipGetByIdentifierQuery.init(
                db=db,
                branch=graphql_context.branch,
                at=graphql_context.at,
                identifiers=ids,
                excluded_namespaces=excluded_namespaces,
                limit=limit,
                offset=offset,
            )

            if "count" in fields:
                response["count"] = await query.count(db=db)

            if not fields:
                return response

            await query.execute(db=db)

            nodes = []
            for peers in query.get_peers():
                nodes.append(
                    {
                        "node": {
                            "id": peers.id,
                            "identifier": peers.identifier,
                            "peers": [
                                {"id": peers.source_id, "kind": peers.source_kind},
                                {"id": peers.destination_id, "kind": peers.destination_kind},
                            ],
                        }
                    }
                )
            response["edges"] = nodes

            return response


Relationship = Field(
    Relationships,
    ids=List(NonNull(String), required=True),
    excluded_namespaces=List(String),
    limit=Int(required=False),
    offset=Int(required=False),
    resolver=Relationships.resolve,
    required=True,
)



import random

from fast_depends import Depends, inject

from .constants import WorkflowTag, WorkflowType
from .models import WorkerPoolDefinition, WorkflowDefinition

INFRAHUB_WORKER_POOL = WorkerPoolDefinition(name="infrahub-worker", description="Default Pool for internal tasks")


ACTION_ADD_NODE_TO_GROUP = WorkflowDefinition(
    name="action-add-node-to-group",
    type=WorkflowType.CORE,
    module="infrahub.actions.tasks",
    function="add_node_to_group",
)

ACTION_RUN_GENERATOR = WorkflowDefinition(
    name="action-run-generator",
    type=WorkflowType.CORE,
    module="infrahub.actions.tasks",
    function="run_generator",
)

ACTION_RUN_GENERATOR_GROUP_EVENT = WorkflowDefinition(
    name="action-run-generator-group-event",
    type=WorkflowType.CORE,
    module="infrahub.actions.tasks",
    function="run_generator_group_event",
)


CONFIGURE_ACTION_RULES = WorkflowDefinition(
    name="configure-action-rules",
    type=WorkflowType.CORE,
    module="infrahub.actions.tasks",
    function="configure_action_rules",
)

TRANSFORM_JINJA2_RENDER = WorkflowDefinition(
    name="transform_render_jinja2_template",
    type=WorkflowType.USER,
    module="infrahub.transformations.tasks",
    function="transform_render_jinja2_template",
)

TRANSFORM_PYTHON_RENDER = WorkflowDefinition(
    name="transform_render_python",
    type=WorkflowType.USER,
    module="infrahub.transformations.tasks",
    function="transform_python",
)

ANONYMOUS_TELEMETRY_SEND = WorkflowDefinition(
    name="anonymous_telemetry_send",
    type=WorkflowType.INTERNAL,
    cron=f"{random.randint(0, 59)} 2 * * *",
    module="infrahub.telemetry.tasks",
    function="send_telemetry_push",
)

SCHEMA_APPLY_MIGRATION = WorkflowDefinition(
    name="schema_apply_migrations",
    type=WorkflowType.CORE,
    module="infrahub.core.migrations.schema.tasks",
    function="schema_apply_migrations",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

SCHEMA_VALIDATE_MIGRATION = WorkflowDefinition(
    name="schema_validate_migrations",
    type=WorkflowType.CORE,
    module="infrahub.core.validators.tasks",
    function="schema_validate_migrations",
)

TRIGGER_ARTIFACT_DEFINITION_GENERATE = WorkflowDefinition(
    name="artifact-definition-generate",
    type=WorkflowType.CORE,
    module="infrahub.git.tasks",
    function="generate_artifact_definition",
)

TRIGGER_GENERATOR_DEFINITION_RUN = WorkflowDefinition(
    name="generator-definition-run",
    type=WorkflowType.CORE,
    module="infrahub.generators.tasks",
    function="run_generator_definition",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

IPAM_RECONCILIATION = WorkflowDefinition(
    name="ipam_reconciliation",
    type=WorkflowType.CORE,
    module="infrahub.core.ipam.tasks",
    function="ipam_reconciliation",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

REMOVE_ADD_NODE_FROM_GROUP = WorkflowDefinition(
    name="action-remove-node-from-group",
    type=WorkflowType.CORE,
    module="infrahub.actions.tasks",
    function="remove_node_from_group",
)

REQUEST_GENERATOR_RUN = WorkflowDefinition(
    name="generator-run",
    type=WorkflowType.USER,
    module="infrahub.generators.tasks",
    function="run_generator",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

RUN_GENERATOR_AS_CHECK = WorkflowDefinition(
    name="run-generator-as-check",
    type=WorkflowType.USER,
    module="infrahub.proposed_change.tasks",
    function="run_generator_as_check",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

REQUEST_GENERATOR_DEFINITION_RUN = WorkflowDefinition(
    name="request-generator-definition-run",
    type=WorkflowType.CORE,
    module="infrahub.generators.tasks",
    function="request_generator_definition_run",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

REQUEST_GENERATOR_DEFINITION_CHECK = WorkflowDefinition(
    name="request-generator-definition-check",
    type=WorkflowType.CORE,
    module="infrahub.proposed_change.tasks",
    function="request_generator_definition_check",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

REQUEST_ARTIFACT_GENERATE = WorkflowDefinition(
    name="artifact-generate",
    type=WorkflowType.CORE,  # NOTE need to check
    module="infrahub.git.tasks",
    function="generate_artifact",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

REQUEST_ARTIFACT_DEFINITION_GENERATE = WorkflowDefinition(
    name="request_artifact_definitions_generate",
    type=WorkflowType.CORE,
    module="infrahub.git.tasks",
    function="generate_request_artifact_definition",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

DIFF_UPDATE = WorkflowDefinition(
    name="diff-update",
    type=WorkflowType.CORE,
    module="infrahub.core.diff.tasks",
    function="update_diff",
)

DIFF_REFRESH = WorkflowDefinition(
    name="diff-refresh",
    type=WorkflowType.CORE,
    module="infrahub.core.diff.tasks",
    function="refresh_diff",
)

DIFF_REFRESH_ALL = WorkflowDefinition(
    name="diff-refresh-all",
    type=WorkflowType.INTERNAL,
    module="infrahub.core.diff.tasks",
    function="refresh_diff_all",
)

GIT_REPOSITORIES_SYNC = WorkflowDefinition(
    name="git_repositories_sync",
    type=WorkflowType.INTERNAL,
    cron="* * * * *",
    module="infrahub.git.tasks",
    function="sync_remote_repositories",
)

GIT_REPOSITORIES_CREATE_BRANCH = WorkflowDefinition(
    name="git-repositories-create-branch",
    type=WorkflowType.CORE,
    module="infrahub.git.tasks",
    function="create_branch",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

GIT_REPOSITORY_ADD = WorkflowDefinition(
    name="git-repository-add-read-write",
    type=WorkflowType.CORE,
    module="infrahub.git.tasks",
    function="add_git_repository",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

GIT_REPOSITORY_ADD_READ_ONLY = WorkflowDefinition(
    name="git-repository-add-read-only",
    type=WorkflowType.CORE,
    module="infrahub.git.tasks",
    function="add_git_repository_read_only",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

GIT_REPOSITORIES_PULL_READ_ONLY = WorkflowDefinition(
    name="git-repository-pull-read-only",
    type=WorkflowType.CORE,
    module="infrahub.git.tasks",
    function="pull_read_only",
)

GIT_REPOSITORIES_MERGE = WorkflowDefinition(
    name="git-repository-merge",
    type=WorkflowType.CORE,
    module="infrahub.git.tasks",
    function="merge_git_repository",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

BRANCH_REBASE = WorkflowDefinition(
    name="branch-rebase",
    type=WorkflowType.CORE,
    module="infrahub.core.branch.tasks",
    function="rebase_branch",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

BRANCH_CREATE = WorkflowDefinition(
    name="create-branch",
    type=WorkflowType.CORE,
    module="infrahub.core.branch.tasks",
    function="create_branch",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

BRANCH_MERGE = WorkflowDefinition(
    name="branch-merge",
    type=WorkflowType.CORE,
    module="infrahub.core.branch.tasks",
    function="merge_branch",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

BRANCH_MERGED = WorkflowDefinition(
    name="branch-merged",
    type=WorkflowType.CORE,
    module="infrahub.branch.tasks",
    function="branch_merged",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

BRANCH_MERGE_POST_PROCESS = WorkflowDefinition(
    name="branch-merge-post-process",
    type=WorkflowType.CORE,
    module="infrahub.core.branch.tasks",
    function="post_process_branch_merge",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

BRANCH_MERGE_MUTATION = WorkflowDefinition(
    name="merge-branch-mutation",
    type=WorkflowType.CORE,
    module="infrahub.graphql.mutations.tasks",
    function="merge_branch_mutation",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

BRANCH_DELETE = WorkflowDefinition(
    name="branch-delete",
    type=WorkflowType.CORE,
    module="infrahub.core.branch.tasks",
    function="delete_branch",
)

BRANCH_VALIDATE = WorkflowDefinition(
    name="branch-validate",
    type=WorkflowType.CORE,
    module="infrahub.core.branch.tasks",
    function="validate_branch",
)

BRANCH_CANCEL_PROPOSED_CHANGES = WorkflowDefinition(
    name="proposed-changes-cancel-branch",
    type=WorkflowType.CORE,
    module="infrahub.proposed_change.tasks",
    function="cancel_proposed_changes_branch",
)

PROPOSED_CHANGE_MERGE = WorkflowDefinition(
    name="proposed-change-merge",
    type=WorkflowType.CORE,
    module="infrahub.proposed_change.tasks",
    function="merge_proposed_change",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

GRAPHQL_QUERY_GROUP_UPDATE = WorkflowDefinition(
    name="graphql-query-group-update",
    type=WorkflowType.CORE,
    module="infrahub.groups.tasks",
    function="update_graphql_query_group",
)

COMPUTED_ATTRIBUTE_PROCESS_JINJA2 = WorkflowDefinition(
    name="computed_attribute_process_jinja2",
    type=WorkflowType.CORE,
    module="infrahub.computed_attribute.tasks",
    function="process_jinja2",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

COMPUTED_ATTRIBUTE_JINJA2_UPDATE_VALUE = WorkflowDefinition(
    name="computed-attribute-jinja2-update-value",
    type=WorkflowType.CORE,
    module="infrahub.computed_attribute.tasks",
    function="computed_attribute_jinja2_update_value",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

TRIGGER_UPDATE_JINJA_COMPUTED_ATTRIBUTES = WorkflowDefinition(
    name="trigger_update_jinja2_computed_attributes",
    type=WorkflowType.CORE,
    module="infrahub.computed_attribute.tasks",
    function="trigger_update_jinja2_computed_attributes",
)

TRIGGER_UPDATE_PYTHON_COMPUTED_ATTRIBUTES = WorkflowDefinition(
    name="trigger_update_python_computed_attributes",
    type=WorkflowType.CORE,
    module="infrahub.computed_attribute.tasks",
    function="trigger_update_python_computed_attributes",
)

COMPUTED_ATTRIBUTE_SETUP_JINJA2 = WorkflowDefinition(
    name="computed-attribute-setup-jinja2",
    type=WorkflowType.CORE,
    module="infrahub.computed_attribute.tasks",
    function="computed_attribute_setup_jinja2",
)

COMPUTED_ATTRIBUTE_SETUP_PYTHON = WorkflowDefinition(
    name="computed-attribute-setup-python",
    type=WorkflowType.CORE,
    module="infrahub.computed_attribute.tasks",
    function="computed_attribute_setup_python",
)

COMPUTED_ATTRIBUTE_PROCESS_TRANSFORM = WorkflowDefinition(
    name="computed_attribute_process_transform",
    type=WorkflowType.USER,
    module="infrahub.computed_attribute.tasks",
    function="process_transform",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

QUERY_COMPUTED_ATTRIBUTE_TRANSFORM_TARGETS = WorkflowDefinition(
    name="query-computed-attribute-transform-targets",
    type=WorkflowType.CORE,
    module="infrahub.computed_attribute.tasks",
    function="query_transform_targets",
)

REQUEST_PROPOSED_CHANGE_DATA_INTEGRITY = WorkflowDefinition(
    name="proposed-changed-data-integrity",
    type=WorkflowType.CORE,
    module="infrahub.proposed_change.tasks",
    function="run_proposed_change_data_integrity_check",
)

REQUEST_PROPOSED_CHANGE_SCHEMA_INTEGRITY = WorkflowDefinition(
    name="proposed-changed-schema-integrity",
    type=WorkflowType.CORE,
    module="infrahub.proposed_change.tasks",
    function="run_proposed_change_schema_integrity_check",
)

REQUEST_PROPOSED_CHANGE_USER_TESTS = WorkflowDefinition(
    name="proposed-changed-user-tests",
    type=WorkflowType.USER,
    module="infrahub.proposed_change.tasks",
    function="run_proposed_change_user_tests",
)

GIT_REPOSITORIES_DIFF_NAMES_ONLY = WorkflowDefinition(
    name="git-repository-diff-names-only",
    type=WorkflowType.INTERNAL,
    module="infrahub.git.tasks",
    function="git_repository_diff_names_only",
)

GIT_REPOSITORIES_IMPORT_OBJECTS = WorkflowDefinition(
    name="git-repository-import-object",
    type=WorkflowType.USER,
    module="infrahub.git.tasks",
    function="import_objects_from_git_repository",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

REQUEST_PROPOSED_CHANGE_PIPELINE = WorkflowDefinition(
    name="proposed-changed-pipeline",
    type=WorkflowType.INTERNAL,
    module="infrahub.proposed_change.tasks",
    function="run_proposed_change_pipeline",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

REQUEST_PROPOSED_CHANGE_REFRESH_ARTIFACTS = WorkflowDefinition(
    name="proposed-changed-refresh-artifacts",
    type=WorkflowType.INTERNAL,
    module="infrahub.proposed_change.tasks",
    function="refresh_artifacts",
)

REQUEST_PROPOSED_CHANGE_RUN_GENERATORS = WorkflowDefinition(
    name="proposed-changed-run-generator",
    type=WorkflowType.INTERNAL,
    module="infrahub.proposed_change.tasks",
    function="run_generators",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

REQUEST_PROPOSED_CHANGE_REPOSITORY_CHECKS = WorkflowDefinition(
    name="proposed-changed-repository-checks",
    type=WorkflowType.INTERNAL,
    module="infrahub.proposed_change.tasks",
    function="repository_checks",
)

REQUEST_ARTIFACT_DEFINITION_CHECK = WorkflowDefinition(
    name="artifacts-generation-validation",
    type=WorkflowType.INTERNAL,
    module="infrahub.proposed_change.tasks",
    function="validate_artifacts_generation",
)

WEBHOOK_PROCESS = WorkflowDefinition(
    name="webhook-process",
    type=WorkflowType.USER,
    module="infrahub.webhook.tasks",
    function="webhook_process",
)

WEBHOOK_CONFIGURE_ONE = WorkflowDefinition(
    name="webhook-setup-automation-one",
    type=WorkflowType.CORE,
    module="infrahub.webhook.tasks",
    function="configure_webhook_one",
)

WEBHOOK_CONFIGURE_ALL = WorkflowDefinition(
    name="webhook-setup-automation-all",
    type=WorkflowType.INTERNAL,
    cron=f"{random.randint(0, 59)} 3 * * *",
    module="infrahub.webhook.tasks",
    function="configure_webhook_all",
)

WEBHOOK_DELETE_AUTOMATION = WorkflowDefinition(
    name="webhook-delete-automation",
    type=WorkflowType.CORE,
    module="infrahub.webhook.tasks",
    function="delete_webhook_automation",
)

GIT_REPOSITORIES_CHECK_ARTIFACT_CREATE = WorkflowDefinition(
    name="git-repository-check-artifact-create",
    type=WorkflowType.USER,
    module="infrahub.artifacts.tasks",
    function="create",
)

GIT_REPOSITORY_USER_CHECKS_DEFINITIONS_TRIGGER = WorkflowDefinition(
    name="git-repository-user-checks-definition-trigger",
    type=WorkflowType.USER,
    module="infrahub.git.tasks",
    function="trigger_repository_user_checks_definitions",
)

GIT_REPOSITORY_USER_CHECK_RUN = WorkflowDefinition(
    name="git-repository-run-user-check",
    type=WorkflowType.USER,
    module="infrahub.git.tasks",
    function="run_user_check",
)

GIT_REPOSITORY_USER_CHECKS_TRIGGER = WorkflowDefinition(
    name="git-repository-trigger-user-checks",
    type=WorkflowType.USER,
    module="infrahub.git.tasks",
    function="trigger_user_checks",
)

GIT_REPOSITORY_INTERNAL_CHECKS_TRIGGER = WorkflowDefinition(
    name="git-repository-trigger-internal-checks",
    type=WorkflowType.USER,
    module="infrahub.git.tasks",
    function="trigger_internal_checks",
)

GIT_REPOSITORY_MERGE_CONFLICTS_CHECKS_RUN = WorkflowDefinition(
    name="git-repository-check-merge-conflict",
    type=WorkflowType.USER,
    module="infrahub.git.tasks",
    function="run_check_merge_conflicts",
)

SCHEMA_UPDATED = WorkflowDefinition(
    name="schema-updated",
    type=WorkflowType.CORE,
    module="infrahub.schema.tasks",
    function="schema_updated",
)


TRIGGER_CONFIGURE_ALL = WorkflowDefinition(
    name="trigger-configure-all",
    type=WorkflowType.CORE,
    module="infrahub.trigger.tasks",
    function="trigger_configure_all",
)


VALIDATE_SCHEMA_NUMBER_POOLS = WorkflowDefinition(
    name="validate-schema-number-pools",
    type=WorkflowType.CORE,
    module="infrahub.pools.tasks",
    function="validate_schema_number_pools",
)


WORKER_POOLS = [INFRAHUB_WORKER_POOL]

WORKFLOWS = [
    ACTION_ADD_NODE_TO_GROUP,
    ACTION_RUN_GENERATOR,
    ACTION_RUN_GENERATOR_GROUP_EVENT,
    ANONYMOUS_TELEMETRY_SEND,
    BRANCH_CANCEL_PROPOSED_CHANGES,
    BRANCH_CREATE,
    BRANCH_DELETE,
    BRANCH_MERGE,
    BRANCH_MERGED,
    BRANCH_MERGE_MUTATION,
    BRANCH_MERGE_POST_PROCESS,
    BRANCH_REBASE,
    BRANCH_VALIDATE,
    COMPUTED_ATTRIBUTE_JINJA2_UPDATE_VALUE,
    COMPUTED_ATTRIBUTE_PROCESS_JINJA2,
    COMPUTED_ATTRIBUTE_PROCESS_TRANSFORM,
    COMPUTED_ATTRIBUTE_SETUP_JINJA2,
    COMPUTED_ATTRIBUTE_SETUP_PYTHON,
    CONFIGURE_ACTION_RULES,
    DIFF_REFRESH,
    DIFF_REFRESH_ALL,
    DIFF_UPDATE,
    GIT_REPOSITORIES_CHECK_ARTIFACT_CREATE,
    GIT_REPOSITORIES_CREATE_BRANCH,
    GIT_REPOSITORIES_DIFF_NAMES_ONLY,
    GIT_REPOSITORIES_IMPORT_OBJECTS,
    GIT_REPOSITORIES_MERGE,
    GIT_REPOSITORIES_PULL_READ_ONLY,
    GIT_REPOSITORIES_SYNC,
    GIT_REPOSITORY_ADD,
    GIT_REPOSITORY_ADD_READ_ONLY,
    GIT_REPOSITORY_INTERNAL_CHECKS_TRIGGER,
    GIT_REPOSITORY_MERGE_CONFLICTS_CHECKS_RUN,
    GIT_REPOSITORY_USER_CHECKS_DEFINITIONS_TRIGGER,
    GIT_REPOSITORY_USER_CHECKS_TRIGGER,
    GIT_REPOSITORY_USER_CHECK_RUN,
    GRAPHQL_QUERY_GROUP_UPDATE,
    IPAM_RECONCILIATION,
    PROPOSED_CHANGE_MERGE,
    QUERY_COMPUTED_ATTRIBUTE_TRANSFORM_TARGETS,
    REMOVE_ADD_NODE_FROM_GROUP,
    REQUEST_ARTIFACT_DEFINITION_CHECK,
    REQUEST_ARTIFACT_DEFINITION_GENERATE,
    REQUEST_ARTIFACT_GENERATE,
    REQUEST_GENERATOR_DEFINITION_CHECK,
    REQUEST_GENERATOR_DEFINITION_RUN,
    REQUEST_GENERATOR_RUN,
    REQUEST_PROPOSED_CHANGE_DATA_INTEGRITY,
    REQUEST_PROPOSED_CHANGE_PIPELINE,
    REQUEST_PROPOSED_CHANGE_REFRESH_ARTIFACTS,
    REQUEST_PROPOSED_CHANGE_REPOSITORY_CHECKS,
    REQUEST_PROPOSED_CHANGE_RUN_GENERATORS,
    REQUEST_PROPOSED_CHANGE_SCHEMA_INTEGRITY,
    REQUEST_PROPOSED_CHANGE_USER_TESTS,
    RUN_GENERATOR_AS_CHECK,
    SCHEMA_APPLY_MIGRATION,
    SCHEMA_UPDATED,
    SCHEMA_VALIDATE_MIGRATION,
    TRANSFORM_JINJA2_RENDER,
    TRANSFORM_PYTHON_RENDER,
    TRIGGER_ARTIFACT_DEFINITION_GENERATE,
    TRIGGER_CONFIGURE_ALL,
    TRIGGER_GENERATOR_DEFINITION_RUN,
    TRIGGER_UPDATE_JINJA_COMPUTED_ATTRIBUTES,
    TRIGGER_UPDATE_PYTHON_COMPUTED_ATTRIBUTES,
    VALIDATE_SCHEMA_NUMBER_POOLS,
    WEBHOOK_CONFIGURE_ALL,
    WEBHOOK_CONFIGURE_ONE,
    WEBHOOK_DELETE_AUTOMATION,
    WEBHOOK_PROCESS,
]


# Use this dependency injection mechanism to easily add new workflows within infrahub-enterprise
def build_workflows_definitions() -> list[WorkflowDefinition]:
    return WORKFLOWS


@inject
def get_workflows(
    workflows: list[WorkflowDefinition] = Depends(build_workflows_definitions),  # noqa: B008
) -> list[WorkflowDefinition]:
    return workflows



from __future__ import annotations

import importlib
import logging
import os
from collections import defaultdict
from csv import DictReader, DictWriter
from datetime import datetime, timezone
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, Sequence

import typer
import ujson
from infrahub_sdk.async_typer import AsyncTyper
from prefect.testing.utilities import prefect_test_harness
from rich import print as rprint
from rich.console import Console
from rich.logging import RichHandler
from rich.table import Table

from infrahub import config
from infrahub.core import registry
from infrahub.core.graph import GRAPH_VERSION
from infrahub.core.graph.constraints import ConstraintManagerBase, ConstraintManagerMemgraph, ConstraintManagerNeo4j
from infrahub.core.graph.index import node_indexes, rel_indexes
from infrahub.core.graph.schema import (
    GRAPH_SCHEMA,
    GraphAttributeProperties,
    GraphNodeProperties,
    GraphRelationshipDefault,
    GraphRelationshipIsPartOf,
    GraphRelationshipProperties,
)
from infrahub.core.initialization import (
    first_time_initialization,
    get_root_node,
    initialization,
    initialize_registry,
)
from infrahub.core.migrations.graph import get_graph_migrations, get_migration_by_number
from infrahub.core.migrations.schema.models import SchemaApplyMigrationData
from infrahub.core.migrations.schema.tasks import schema_apply_migrations
from infrahub.core.schema import SchemaRoot, core_models, internal_schema
from infrahub.core.schema.definitions.deprecated import deprecated_models
from infrahub.core.schema.manager import SchemaManager
from infrahub.core.utils import delete_all_nodes
from infrahub.core.validators.models.validate_migration import SchemaValidateMigrationData
from infrahub.core.validators.tasks import schema_validate_migrations
from infrahub.database import DatabaseType
from infrahub.database.memgraph import IndexManagerMemgraph
from infrahub.database.neo4j import IndexManagerNeo4j
from infrahub.log import get_logger

from .constants import ERROR_BADGE, FAILED_BADGE, SUCCESS_BADGE
from .db_commands.check_inheritance import check_inheritance
from .patch import patch_app


def get_timestamp_string() -> str:
    """Generate a timestamp string in the format YYYYMMDD-HHMMSS."""
    return datetime.now(tz=timezone.utc).strftime("%Y%m%d-%H%M%S")


if TYPE_CHECKING:
    from infrahub.cli.context import CliContext
    from infrahub.core.migrations.shared import ArbitraryMigration, GraphMigration, InternalSchemaMigration
    from infrahub.database import InfrahubDatabase
    from infrahub.database.index import IndexManagerBase

app = AsyncTyper()
app.add_typer(patch_app, name="patch")

PERMISSIONS_AVAILABLE = ["read", "write", "admin"]


class ConstraintAction(str, Enum):
    SHOW = "show"
    ADD = "add"
    DROP = "drop"


class IndexAction(str, Enum):
    SHOW = "show"
    ADD = "add"
    DROP = "drop"


@app.callback()
def callback() -> None:
    """
    Manage the graph in the database.
    """


@app.command()
async def init(
    ctx: typer.Context,
    config_file: str = typer.Option(
        "infrahub.toml", envvar="INFRAHUB_CONFIG", help="Location of the configuration file to use for Infrahub"
    ),
) -> None:
    """Erase the content of the database and initialize it with the core schema."""

    log = get_logger()

    # --------------------------------------------------
    # CLEANUP
    #  - For now we delete everything in the database
    #   TODO, if possible try to implement this in an idempotent way
    # --------------------------------------------------

    logging.getLogger("neo4j").setLevel(logging.ERROR)
    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)
    async with dbdriver.start_transaction() as db:
        log.info("Delete All Nodes")
        await delete_all_nodes(db=db)
        await first_time_initialization(db=db)

    await dbdriver.close()


@app.command()
async def load_test_data(
    ctx: typer.Context,
    config_file: str = typer.Option(
        "infrahub.toml", envvar="INFRAHUB_CONFIG", help="Location of the configuration file to use for Infrahub"
    ),
    dataset: str = "dataset01",
) -> None:
    """Load test data into the database from the `test_data` directory."""

    logging.getLogger("neo4j").setLevel(logging.ERROR)
    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)

    async with dbdriver.start_session() as db:
        await initialization(db=db)

        log_level = "DEBUG"

        FORMAT = "%(message)s"
        logging.basicConfig(level=log_level, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])
        logging.getLogger("infrahub")

        dataset_module = importlib.import_module(f"infrahub.test_data.{dataset}")
        await dataset_module.load_data(db=db)

    await dbdriver.close()


@app.command(name="migrate")
async def migrate_cmd(
    ctx: typer.Context,
    check: bool = typer.Option(False, help="Check the state of the database without applying the migrations."),
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
    migration_number: int | None = typer.Option(None, help="Apply a specific migration by number"),
) -> None:
    """Check the current format of the internal graph and apply the necessary migrations"""
    logging.getLogger("infrahub").setLevel(logging.WARNING)
    logging.getLogger("neo4j").setLevel(logging.ERROR)
    logging.getLogger("prefect").setLevel(logging.ERROR)

    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)

    await migrate_database(db=dbdriver, initialize=True, check=check, migration_number=migration_number)

    await dbdriver.close()


@app.command(name="check-inheritance")
async def check_inheritance_cmd(
    ctx: typer.Context,
    fix: bool = typer.Option(False, help="Fix the inheritance of any invalid nodes."),
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    """Check the database for any vertices with incorrect inheritance"""
    logging.getLogger("infrahub").setLevel(logging.WARNING)
    logging.getLogger("neo4j").setLevel(logging.ERROR)
    logging.getLogger("prefect").setLevel(logging.ERROR)

    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)
    await initialize_registry(db=dbdriver)

    success = await check_inheritance(db=dbdriver, fix=fix)
    if not success:
        raise typer.Exit(code=1)

    await dbdriver.close()


@app.command(name="update-core-schema")
async def update_core_schema_cmd(
    ctx: typer.Context,
    debug: bool = typer.Option(False, help="Enable advanced logging and troubleshooting"),
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    """Check the current format of the internal graph and apply the necessary migrations"""
    logging.getLogger("infrahub").setLevel(logging.WARNING)
    logging.getLogger("neo4j").setLevel(logging.ERROR)
    logging.getLogger("prefect").setLevel(logging.ERROR)
    os.environ["PREFECT_SERVER_ANALYTICS_ENABLED"] = "false"

    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)

    with prefect_test_harness():
        await update_core_schema(db=dbdriver, initialize=True, debug=debug)

    await dbdriver.close()


@app.command()
async def constraint(
    ctx: typer.Context,
    action: ConstraintAction = typer.Argument(ConstraintAction.SHOW),
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    """Manage Database Constraints"""
    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)

    manager: ConstraintManagerBase | None = None
    if dbdriver.db_type == DatabaseType.NEO4J:
        manager = ConstraintManagerNeo4j.from_graph_schema(db=dbdriver, schema=GRAPH_SCHEMA)
    elif dbdriver.db_type == DatabaseType.MEMGRAPH:
        manager = ConstraintManagerMemgraph.from_graph_schema(db=dbdriver, schema=GRAPH_SCHEMA)
    else:
        print(f"Database type not supported : {dbdriver.db_type}")
        raise typer.Exit(1)

    if action == ConstraintAction.ADD:
        await manager.add()
    elif action == ConstraintAction.DROP:
        await manager.drop()

    constraints = await manager.list()

    console = Console()

    table = Table(title="Database Constraints")

    table.add_column("Name", justify="right", style="cyan", no_wrap=True)
    table.add_column("Label")
    table.add_column("Property")

    for item in constraints:
        table.add_row(item.item_name, item.item_label, item.property)

    console.print(table)

    await dbdriver.close()


@app.command()
async def index(
    ctx: typer.Context,
    action: IndexAction = typer.Argument(IndexAction.SHOW),
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    """Manage Database Indexes"""
    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)
    if dbdriver.db_type is DatabaseType.MEMGRAPH:
        index_manager: IndexManagerBase = IndexManagerMemgraph(db=dbdriver)
    index_manager = IndexManagerNeo4j(db=dbdriver)

    index_manager.init(nodes=node_indexes, rels=rel_indexes)

    if action == IndexAction.ADD:
        await index_manager.add()
    elif action == IndexAction.DROP:
        await index_manager.drop()

    indexes = await index_manager.list()

    console = Console()

    table = Table(title="Database Indexes")

    table.add_column("Name", justify="right", style="cyan", no_wrap=True)
    table.add_column("Label")
    table.add_column("Property")
    table.add_column("Type")
    table.add_column("Entity Type")

    for item in indexes:
        table.add_row(
            item.name, item.label, ", ".join(item.properties), item.type.value.upper(), item.entity_type.value.upper()
        )

    console.print(table)

    await dbdriver.close()


async def migrate_database(
    db: InfrahubDatabase, initialize: bool = False, check: bool = False, migration_number: int | str | None = None
) -> bool:
    """Apply the latest migrations to the database, this function will print the status directly in the console.

    Returns a boolean indicating whether a migration failed or if all migrations succeeded.

    Args:
        db: The database object.
        check: If True, the function will only check the status of the database and not apply the migrations. Defaults to False.
        migration_number: If provided, the function will only apply the migration with the given number. Defaults to None.
    """
    rprint("Checking current state of the Database")

    if initialize:
        await initialize_registry(db=db)

    root_node = await get_root_node(db=db)
    if migration_number:
        migration = get_migration_by_number(migration_number)
        migrations: Sequence[GraphMigration | InternalSchemaMigration | ArbitraryMigration] = [migration]
        if check:
            if root_node.graph_version > migration.minimum_version:
                rprint(
                    f"Migration {migration_number} already applied. To apply again, run the command without the --check flag."
                )
                return True
            rprint(
                f"Migration {migration_number} needs to be applied. Run `infrahub db migrate` to apply all outstanding migrations."
            )
            return False
    else:
        migrations = await get_graph_migrations(root=root_node)
        if not migrations:
            rprint(f"Database up-to-date (v{root_node.graph_version}), no migration to execute.")
            return True

        rprint(
            f"Database needs to be updated (v{root_node.graph_version} -> v{GRAPH_VERSION}), {len(migrations)} migrations pending"
        )

    if check:
        return True

    for migration in migrations:
        execution_result = await migration.execute(db=db)
        validation_result = None

        if execution_result.success:
            validation_result = await migration.validate_migration(db=db)
            if validation_result.success:
                rprint(f"Migration: {migration.name} {SUCCESS_BADGE}")
                root_node.graph_version = migration.minimum_version + 1
                await root_node.save(db=db)

        if not execution_result.success or (validation_result and not validation_result.success):
            rprint(f"Migration: {migration.name} {FAILED_BADGE}")
            for error in execution_result.errors:
                rprint(f"  {error}")
            if validation_result and not validation_result.success:
                for error in validation_result.errors:
                    rprint(f"  {error}")
            return False

    return True


async def initialize_internal_schema() -> None:
    registry.schema = SchemaManager()
    schema = SchemaRoot(**internal_schema)
    registry.schema.register_schema(schema=schema)


async def update_core_schema(db: InfrahubDatabase, initialize: bool = True, debug: bool = False) -> None:
    """Update the core schema of Infrahub to the latest version"""
    # ----------------------------------------------------------
    # Initialize Schema and Registry
    # ----------------------------------------------------------
    if initialize:
        await initialize_registry(db=db)
        await initialize_internal_schema()

    default_branch = registry.get_branch_from_registry(branch=registry.default_branch)

    # ----------------------------------------------------------
    # Load Current Schema from the database
    # ----------------------------------------------------------
    schema_default_branch = await registry.schema.load_schema_from_db(db=db, branch=default_branch)
    registry.schema.set_schema_branch(name=default_branch.name, schema=schema_default_branch)
    branch_schema = registry.schema.get_schema_branch(name=registry.default_branch)

    candidate_schema = branch_schema.duplicate()
    candidate_schema.load_schema(schema=SchemaRoot(**internal_schema))
    candidate_schema.load_schema(schema=SchemaRoot(**core_models))
    candidate_schema.load_schema(schema=SchemaRoot(**deprecated_models))
    candidate_schema.process()

    schema_diff = branch_schema.diff(other=candidate_schema)
    branch_schema.validate_node_deletions(diff=schema_diff)
    result = branch_schema.validate_update(other=candidate_schema, diff=schema_diff, enforce_update_support=False)
    if result.errors:
        rprint(f"{ERROR_BADGE} | Unable to update the schema, due to failed validations")
        for error in result.errors:
            rprint(error.to_string())
        raise typer.Exit(1)

    if not result.diff.all:
        rprint("Core Schema Up to date, nothing to update")
        return

    rprint("Core Schema has diff, will need to be updated")
    if debug:
        result.diff.print()

    # ----------------------------------------------------------
    # Validate if the new schema is valid with the content of the database
    # ----------------------------------------------------------
    validate_migration_data = SchemaValidateMigrationData(
        branch=default_branch,
        schema_branch=candidate_schema,
        constraints=result.constraints,
    )
    responses = await schema_validate_migrations(message=validate_migration_data)
    error_messages = [violation.message for response in responses for violation in response.violations]
    if error_messages:
        rprint(f"{ERROR_BADGE} | Unable to update the schema, due to failed validations")
        for message in error_messages:
            rprint(message)
        raise typer.Exit(1)

    # ----------------------------------------------------------
    # Update the schema
    # ----------------------------------------------------------
    origin_schema = branch_schema.duplicate()

    # Update the internal schema
    schema_default_branch.load_schema(schema=SchemaRoot(**internal_schema))
    schema_default_branch.process()
    registry.schema.set_schema_branch(name=default_branch.name, schema=schema_default_branch)

    async with db.start_transaction() as dbt:
        await registry.schema.update_schema_branch(
            schema=candidate_schema,
            db=dbt,
            branch=default_branch.name,
            diff=result.diff,
            limit=result.diff.all,
            update_db=True,
        )
        default_branch.update_schema_hash()
        rprint("The Core Schema has been updated, make sure to rebase any open branches after the upgrade")
        if debug:
            rprint(f"New schema hash: {default_branch.active_schema_hash.main}")
        await default_branch.save(db=dbt)

    # ----------------------------------------------------------
    # Run the migrations
    # ----------------------------------------------------------
    apply_migration_data = SchemaApplyMigrationData(
        branch=default_branch,
        new_schema=candidate_schema,
        previous_schema=origin_schema,
        migrations=result.migrations,
    )
    migration_error_msgs = await schema_apply_migrations(message=apply_migration_data)

    if migration_error_msgs:
        rprint(f"{ERROR_BADGE} | Some error(s) happened while running the schema migrations")
        for message in migration_error_msgs:
            rprint(message)
        raise typer.Exit(1)


@app.command(name="selected-export")
async def selected_export_cmd(
    ctx: typer.Context,
    kinds: list[str] = typer.Option([], help="Node kinds to export"),  # noqa: B008
    uuids: list[str] = typer.Option([], help="UUIDs of nodes to export"),  # noqa: B008
    query_limit: int = typer.Option(1000, help="Maximum batch size of export query"),
    export_dir: Path = typer.Option(Path("infrahub-exports"), help="Path of directory to save exports"),  # noqa: B008
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    """Export database structure of selected nodes from the database without any actual data"""
    logging.getLogger("infrahub").setLevel(logging.WARNING)
    logging.getLogger("neo4j").setLevel(logging.ERROR)
    logging.getLogger("prefect").setLevel(logging.ERROR)

    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)

    await selected_export(db=dbdriver, kinds=kinds, uuids=uuids, export_dir=export_dir, query_limit=query_limit)

    await dbdriver.close()


async def selected_export(
    db: InfrahubDatabase,
    kinds: list[str],
    uuids: list[str],
    export_dir: Path,
    query_limit: int = 1000,
) -> Path:
    query = """
// --------------
// filter nodes
// --------------
MATCH (n:Node)
WHERE ($kinds IS NULL OR size($kinds) = 0 OR any(l IN labels(n) WHERE l in $kinds))
AND ($uuids IS NULL OR size($uuids) = 0 OR n.uuid IN $uuids)
WITH n
// --------------
// pagination
// --------------
ORDER BY %(id_func)s(n)
SKIP toInteger($offset)
LIMIT toInteger($limit)
CALL (n) {
    // --------------
    // get all the nodes and edges linked to this node up to 2 steps away, excluding IS_PART_OF
    // --------------
    MATCH (n)-[r1]-(v1)-[r2]-(v2)
    WHERE type(r1) <> "IS_PART_OF"
    WITH collect([v1, v2]) AS vertex_pairs, collect([r1, r2]) AS edge_pairs
    WITH reduce(
        vertices = [], v_pair IN vertex_pairs |
        CASE
            WHEN NOT v_pair[0] IN vertices AND NOT v_pair[1] IN vertices THEN vertices + v_pair
            WHEN NOT v_pair[0] IN vertices THEN vertices + [v_pair[0]]
            WHEN NOT v_pair[1] IN vertices THEN vertices + [v_pair[1]]
            ELSE vertices
        END
    ) AS vertices,
    reduce(
        edges = [], e_pair IN edge_pairs |
        CASE
            WHEN NOT e_pair[0] IN edges AND NOT e_pair[1] IN edges THEN edges + e_pair
            WHEN NOT e_pair[0] IN edges THEN edges + [e_pair[0]]
            WHEN NOT e_pair[1] IN edges THEN edges + [e_pair[1]]
            ELSE edges
        END
    ) AS edges
    RETURN vertices, edges
}
// --------------
// include the root and IS_PART_OF edges
// --------------
OPTIONAL MATCH (n)-[root_edge:IS_PART_OF]->(root:Root)
WITH n, vertices, edges, root, collect(root_edge) AS root_edges
WITH n, edges + root_edges AS edges, CASE
    WHEN root IS NOT NULL THEN vertices + [n, root]
    ELSE vertices + [n]
END AS vertices
RETURN vertices, edges
    """ % {"id_func": db.get_id_function_name()}
    timestamp_str = get_timestamp_string()
    export_dir /= Path(f"export-{timestamp_str}")
    if not export_dir.exists():
        export_dir.mkdir(parents=True)
    vertex_path = export_dir / Path("vertices.csv")
    vertex_path.touch(exist_ok=True)
    edge_path = export_dir / Path("edges.csv")
    edge_path.touch(exist_ok=True)

    graph_node_schemas = [GraphNodeProperties, GraphRelationshipProperties, GraphAttributeProperties]
    graph_vertex_properties = set()
    for graph_schema in graph_node_schemas:
        for field_name, field_info in graph_schema.model_fields.items():
            property_name = field_info.alias or field_name
            graph_vertex_properties.add(property_name)

    graph_edge_schemas = [GraphRelationshipIsPartOf, GraphRelationshipDefault]
    graph_edge_properties = set()
    for graph_schema in graph_edge_schemas:
        for field_name, field_info in graph_schema.model_fields.items():
            property_name = field_info.alias or field_name
            graph_edge_properties.add(property_name)

    all_db_ids: set[str] = set()
    has_more_data = True
    limit = query_limit
    offset = 0

    with vertex_path.open(mode="w") as vertex_file, edge_path.open(mode="w") as edge_file:
        vertex_field_names = ["db_id", "labels"] + sorted(graph_vertex_properties)
        vertex_csv_writer = DictWriter(vertex_file, fieldnames=vertex_field_names)
        vertex_csv_writer.writeheader()
        edge_field_names = ["db_id", "edge_type", "start_node_id", "end_node_id"] + sorted(graph_edge_properties)
        edge_csv_writer = DictWriter(edge_file, fieldnames=edge_field_names)
        edge_csv_writer.writeheader()

        while has_more_data:
            rprint("Retrieving batch of vertices and edges...", end="")
            results = await db.execute_query(
                query=query,
                params={"kinds": kinds, "uuids": uuids, "limit": limit, "offset": offset},
            )
            rprint("done. ", end="")
            has_more_data = len(results) >= limit
            offset += limit

            rprint("Writing batch to export files...", end="")
            for result in results:
                vertices = result.get("vertices")
                for vertex in vertices:
                    if vertex.element_id in all_db_ids:
                        continue
                    serial_vertex = {
                        "db_id": vertex.element_id,
                        "labels": ujson.dumps(list(vertex.labels)),
                    }
                    for property_name in graph_vertex_properties:
                        if value := vertex.get(property_name):
                            serial_vertex[property_name] = value
                    vertex_csv_writer.writerow(serial_vertex)
                    all_db_ids.add(vertex.element_id)

                edges = result.get("edges")
                for edge in edges:
                    if edge.element_id in all_db_ids:
                        continue
                    serial_edge = {
                        "db_id": edge.element_id,
                        "edge_type": edge.type,
                        "start_node_id": edge.start_node.element_id,
                        "end_node_id": edge.end_node.element_id,
                    }
                    for property_name in graph_edge_properties:
                        if value := edge.get(property_name):
                            serial_edge[property_name] = value
                    edge_csv_writer.writerow(serial_edge)
                    all_db_ids.add(edge.element_id)
            rprint("done.")

    rprint(f"{SUCCESS_BADGE} Export complete")
    rprint(f"Export directory is here: {export_dir.absolute()}")
    return export_dir


@app.command(name="load-export", hidden=True)
async def load_export_cmd(
    ctx: typer.Context,
    export_dir: Path = typer.Argument(help="Path to export directory"),
    query_limit: int = typer.Option(1000, help="Maximum batch size of import query"),
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    """
    Cannot be used for backup/restore functionality.
    Loads an anonymized export into Neo4j.
    Only used for analysis of output of the selected-export command.
    """
    logging.getLogger("infrahub").setLevel(logging.WARNING)
    logging.getLogger("neo4j").setLevel(logging.ERROR)
    logging.getLogger("prefect").setLevel(logging.ERROR)

    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)

    await load_export(db=dbdriver, export_dir=export_dir, query_limit=query_limit)

    await dbdriver.close()


async def load_vertices(
    db: InfrahubDatabase, vertex_labels: list[str], vertex_dicts: list[dict[str, str | int | bool | None]]
) -> None:
    vertex_import_query = """
UNWIND $vertices AS vertex
CREATE (v:ImportNode:%(node_labels)s {db_id: vertex.db_id})
SET v = vertex
    """ % {"node_labels": ":".join(vertex_labels)}
    rprint(f"Loading {len(vertex_dicts)} {vertex_labels} nodes...", end="")
    await db.execute_query(query=vertex_import_query, params={"vertices": vertex_dicts})
    rprint("done")


async def load_edges(
    db: InfrahubDatabase, edge_type: str, edge_dicts: list[dict[str, str | int | bool | None]]
) -> None:
    edges_import_query = """
UNWIND $edges AS edge
MATCH (a:ImportNode) WHERE a.db_id = toString(edge.start_node_id)
MATCH (b:ImportNode) WHERE b.db_id = toString(edge.end_node_id)
CREATE (a)-[e:%(edge_type)s]->(b)
SET e = edge.properties
    """ % {"edge_type": edge_type}
    rprint(f"Loading {len(edge_dicts)} {edge_type} edges...", end="")
    await db.execute_query(query=edges_import_query, params={"edges": edge_dicts})
    rprint("done")


async def load_export(db: InfrahubDatabase, export_dir: Path, query_limit: int = 1000) -> None:
    if not export_dir.exists():
        rprint(f"{ERROR_BADGE} {export_dir} does not exist")
        raise typer.Exit(1)
    if not export_dir.is_dir():
        rprint(f"{ERROR_BADGE} {export_dir} is not a directory")
        raise typer.Exit(1)
    vertex_file: Path | None = None
    edge_file: Path | None = None

    for export_file in export_dir.glob("*.csv"):
        if export_file.name == "vertices.csv":
            vertex_file = export_file
        elif export_file.name == "edges.csv":
            edge_file = export_file
    if not vertex_file or not vertex_file.exists() or not vertex_file.is_file():
        rprint(f"{ERROR_BADGE} File 'vertices.csv' does not exist in the export directory")
        raise typer.Exit(1)
    if not edge_file or not edge_file.exists() or not edge_file.is_file():
        rprint(f"{ERROR_BADGE} File 'edges.csv' does not exist in the export directory")
        raise typer.Exit(1)

    # index massively improves time required to load a large export
    create_index_query = "CREATE RANGE INDEX import_node_db_id IF NOT EXISTS FOR (v:ImportNode) ON (v.db_id)"
    await db.execute_query(query=create_index_query)

    rprint("Loading vertices...")
    vertices_by_labels_map: dict[frozenset[str], list[dict[str, Any]]] = defaultdict(list)
    with vertex_file.open() as file:
        csv_reader = DictReader(file)
        for vertex_row in csv_reader:
            labels = frozenset(ujson.loads(vertex_row["labels"]))
            cleaned_row = {k: v for k, v in vertex_row.items() if k != "labels" and v}
            vertices_by_labels_map[labels].append(cleaned_row)
            # once a group of vertices meets the query_limit, save them to the database and delete them from memory
            if len(vertices_by_labels_map[labels]) >= query_limit:
                await load_vertices(db=db, vertex_labels=list(labels), vertex_dicts=vertices_by_labels_map[labels])
                vertices_by_labels_map[labels] = []

        for labels, vertex_rows in vertices_by_labels_map.items():
            await load_vertices(db=db, vertex_labels=list(labels), vertex_dicts=vertex_rows)
    rprint("Vertices loaded")

    rprint("Loading edges...")
    edges_by_type_map: dict[str, list[dict[str, Any]]] = defaultdict(list)
    with edge_file.open() as file:
        csv_reader = DictReader(file)
        for edge_row in csv_reader:
            edge_type = edge_row["edge_type"]

            edge_properties = {}
            edge_dict = {}
            for k, v in edge_row.items():
                if k == "edge_type" or not v:
                    continue
                if k in ["start_node_id", "end_node_id"]:
                    edge_dict[k] = v
                else:
                    edge_properties[k] = v
            edge_dict["properties"] = edge_properties
            edges_by_type_map[edge_type].append(edge_dict)
            # once a group of edges meets the query_limit, save them to the database and delete them from memory
            if len(edges_by_type_map[edge_type]) >= query_limit:
                await load_edges(db=db, edge_type=edge_type, edge_dicts=edges_by_type_map[edge_type])
                edges_by_type_map[edge_type] = []

        for edge_type, edge_dicts in edges_by_type_map.items():
            await load_edges(db=db, edge_type=edge_type, edge_dicts=edge_dicts)
    rprint("Edges loaded")
    rprint(f"{SUCCESS_BADGE} Export loaded")


@app.command(name="check")
async def check_cmd(
    ctx: typer.Context,
    output_dir: Path = typer.Option(  # noqa: B008
        None, help="Directory to save detailed check results (defaults to infrahub_db_check_YYYYMMDD-HHMMSS)"
    ),
    config_file: str = typer.Option(
        "infrahub.toml", envvar="INFRAHUB_CONFIG", help="Location of the configuration file to use for Infrahub"
    ),
) -> None:
    """Run database sanity checks and output the results to the CSV files."""
    logging.getLogger("infrahub").setLevel(logging.WARNING)
    logging.getLogger("neo4j").setLevel(logging.ERROR)
    logging.getLogger("prefect").setLevel(logging.ERROR)

    config.load_and_exit(config_file_name=config_file)

    # Create output directory if not specified
    if output_dir is None:
        timestamp_str = get_timestamp_string()
        output_dir = Path(f"infrahub_db_check_{timestamp_str}")

    if not output_dir.exists():
        output_dir.mkdir(parents=True)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)

    await run_database_checks(db=dbdriver, output_dir=output_dir)

    await dbdriver.close()


async def run_database_checks(db: InfrahubDatabase, output_dir: Path) -> None:
    """Run a series of database health checks and output the results to the terminal.

    Args:
        db: The database object.
        output_dir: Directory to save detailed check results.
    """
    rprint("Running database health checks...")

    # Check 1: Duplicate active relationships
    rprint("\n[bold cyan]Check 1: Duplicate Active Relationships[/bold cyan]")
    duplicate_active_rels_query = """
    MATCH (a:Node)-[e1:IS_RELATED {status: "active"}]-(r:Relationship)-[e2:IS_RELATED {branch: e1.branch, status: "active"}]-(b:Node)
    WHERE a.uuid < b.uuid
    AND e1.to IS NULL
    AND e2.to IS NULL
    WITH DISTINCT a.uuid AS a_uuid,
        b.uuid AS b_uuid,
        r.name AS r_name,
        e1.branch AS branch,
        CASE
        WHEN startNode(e1) = a AND startNode(e2) = r THEN "out"
            WHEN startNode(e1) = r AND startNode(e2) = b THEN "in"
            ELSE "bidir"
        END AS direction,
        count(*) AS num_paths,
        collect(DISTINCT a.kind) AS a_kinds,
        collect(DISTINCT b.kind) AS b_kinds
    WHERE num_paths > 1
    RETURN a_uuid, a_kinds, b_uuid, b_kinds, r_name, branch, direction, num_paths
    """

    results = await db.execute_query(query=duplicate_active_rels_query)
    if results:
        rprint(f"[red]Found {len(results)} duplicate active relationships[/red]")
        # Write detailed results to file
        output_file = output_dir / "duplicate_active_relationships.csv"
        with output_file.open(mode="w", newline="") as f:
            writer = DictWriter(
                f, fieldnames=["a_uuid", "a_kinds", "b_uuid", "b_kinds", "r_name", "branch", "direction", "num_paths"]
            )
            writer.writeheader()
            for result in results:
                writer.writerow(dict(result))
        rprint(f"  Detailed results written to: {output_file}")
    else:
        rprint(f"{SUCCESS_BADGE} No duplicate active relationships found")

    # Check 2: Duplicated relationship nodes
    rprint("\n[bold cyan]Check 2: Duplicated Relationship Nodes[/bold cyan]")
    duplicate_rel_nodes_query = """
    MATCH (r:Relationship)
    WITH r.uuid AS r_uuid, COUNT(*) AS num_rels
    WHERE num_rels > 1
    MATCH (n:Node)-[:IS_RELATED]-(r:Relationship {uuid: r_uuid})
    WITH DISTINCT r_uuid, n.uuid AS n_uuid, n.kind AS n_kind
    WITH r_uuid, collect([n_uuid, n_kind]) AS node_details
    RETURN r_uuid, node_details
    """

    results = await db.execute_query(query=duplicate_rel_nodes_query)
    if results:
        rprint(f"[red]Found {len(results)} duplicated relationship nodes[/red]")
        # Write detailed results to file
        output_file = output_dir / "duplicated_relationship_nodes.csv"
        with output_file.open(mode="w", newline="") as f:
            writer = DictWriter(f, fieldnames=["r_uuid", "node_details"])
            writer.writeheader()
            for result in results:
                writer.writerow(dict(result))
        rprint(f"  Detailed results written to: {output_file}")
    else:
        rprint(f"{SUCCESS_BADGE} No duplicated relationship nodes found")

    # Check 3: Duplicated edges
    rprint("\n[bold cyan]Check 3: Duplicated Edges[/bold cyan]")
    duplicate_edges_query = """
    MATCH (a)
    CALL (a) {
        MATCH (a)-[e]->(b)
        WHERE elementId(a) < elementId(b)
        WITH DISTINCT a, b, type(e) AS e_type, count(*) AS total_num_edges
        WHERE total_num_edges > 1
        MATCH (a)-[e]->(b)
        WHERE type(e) = e_type
        WITH
            elementId(a) AS a_id,
            labels(a) AS a_labels,
            elementId(b) AS b_id,
            labels(b) AS b_labels,
            type(e) AS e_type,
            e.branch AS branch,
            e.status AS status,
            e.from AS time,
            collect(e) AS edges
        WITH a_id, a_labels, b_id, b_labels, e_type, branch, status, time, size(edges) AS num_edges
        WHERE num_edges > 1
        WITH a_id, a_labels, b_id, b_labels, e_type, branch, status, time, num_edges
        RETURN a_id, a_labels, b_id, b_labels, e_type, branch, status, time, num_edges
    }
    RETURN a_id, a_labels, b_id, b_labels, e_type, branch, status, time, num_edges
    """

    results = await db.execute_query(query=duplicate_edges_query)
    if results:
        rprint(f"[red]Found {len(results)} sets of duplicated edges[/red]")
        # Write detailed results to file
        output_file = output_dir / "duplicated_edges.csv"
        with output_file.open(mode="w", newline="") as f:
            writer = DictWriter(
                f,
                fieldnames=["a_id", "a_labels", "b_id", "b_labels", "e_type", "branch", "status", "time", "num_edges"],
            )
            writer.writeheader()
            for result in results:
                writer.writerow(dict(result))
        rprint(f"  Detailed results written to: {output_file}")
    else:
        rprint(f"{SUCCESS_BADGE} No duplicated edges found")

    # Check 4: Orphaned Relationships
    rprint("\n[bold cyan]Check 4: Orphaned Relationships[/bold cyan]")
    orphaned_rels_query = """
    MATCH (r:Relationship)-[:IS_RELATED]-(peer:Node)
    WITH DISTINCT r, peer
    WITH r, count(*) AS num_peers
    WHERE num_peers < 2
    MATCH (r)-[e:IS_RELATED]-(peer:Node)
    RETURN DISTINCT
        r.name AS r_name,
        e.branch AS branch,
        e.status AS status,
        e.from AS from_time,
        e.to AS to_time,
        peer.uuid AS peer_uuid,
        peer.kind AS peer_kind
    """
    results = await db.execute_query(query=orphaned_rels_query)
    if results:
        rprint(f"[red]Found {len(results)} orphaned Relationships[/red]")
        # Write detailed results to file
        output_file = output_dir / "orphaned_relationships.csv"
        with output_file.open(mode="w", newline="") as f:
            writer = DictWriter(
                f,
                fieldnames=["r_name", "branch", "status", "from_time", "to_time", "peer_uuid", "peer_kind"],
            )
            writer.writeheader()
            for result in results:
                writer.writerow(dict(result))
        rprint(f"  Detailed results written to: {output_file}")
    else:
        rprint(f"{SUCCESS_BADGE} No orphaned relationships found")

    rprint(f"\n{SUCCESS_BADGE} Database health checks completed")
    rprint(f"Detailed results saved to: {output_dir.absolute()}")
